from kopf import BaseRegistry, GlobalRegistry, SimpleRegistry


def test_creation_of_simple():
    registry = SimpleRegistry()
    assert isinstance(registry, BaseRegistry)
    assert isinstance(registry, SimpleRegistry)


def test_creation_of_global():
    registry = GlobalRegistry()
    assert isinstance(registry, BaseRegistry)
    assert isinstance(registry, GlobalRegistry)



# Original test-file: tests/basic-structs/test_handlers.py
import pytest

from kopf.structs.handlers import ActivityHandler, ResourceChangingHandler


def test_activity_handler_with_deprecated_cooldown_instead_of_backoff(mocker):
    fn = mocker.Mock()
    id = mocker.Mock()
    errors = mocker.Mock()
    timeout = mocker.Mock()
    retries = mocker.Mock()
    backoff = mocker.Mock()
    activity = mocker.Mock()

    with pytest.deprecated_call(match=r"use backoff="):
        handler = ActivityHandler(
            fn=fn,
            id=id,
            errors=errors,
            timeout=timeout,
            retries=retries,
            backoff=None,
            cooldown=backoff,  # deprecated, but still required
            activity=activity,
        )

    assert handler.fn is fn
    assert handler.id is id
    assert handler.errors is errors
    assert handler.timeout is timeout
    assert handler.retries is retries
    assert handler.backoff is backoff
    assert handler.activity is activity

    with pytest.deprecated_call(match=r"use handler.backoff"):
        assert handler.cooldown is backoff


def test_resource_handler_with_deprecated_cooldown_instead_of_backoff(mocker):
    fn = mocker.Mock()
    id = mocker.Mock()
    reason = mocker.Mock()
    field = mocker.Mock()
    errors = mocker.Mock()
    timeout = mocker.Mock()
    retries = mocker.Mock()
    backoff = mocker.Mock()
    initial = mocker.Mock()
    deleted = mocker.Mock()
    labels = mocker.Mock()
    annotations = mocker.Mock()
    when = mocker.Mock()
    requires_finalizer = mocker.Mock()

    with pytest.deprecated_call(match=r"use backoff="):
        handler = ResourceChangingHandler(
            fn=fn,
            id=id,
            reason=reason,
            field=field,
            errors=errors,
            timeout=timeout,
            retries=retries,
            backoff=None,
            cooldown=backoff,  # deprecated, but still required
            initial=initial,
            deleted=deleted,
            labels=labels,
            annotations=annotations,
            when=when,
            requires_finalizer=requires_finalizer,
        )

    assert handler.fn is fn
    assert handler.id is id
    assert handler.reason is reason
    assert handler.field is field
    assert handler.errors is errors
    assert handler.timeout is timeout
    assert handler.retries is retries
    assert handler.backoff is backoff
    assert handler.initial is initial
    assert handler.deleted is deleted
    assert handler.labels is labels
    assert handler.annotations is annotations
    assert handler.when is when
    assert handler.requires_finalizer is requires_finalizer



import pytest

import kopf
from kopf.structs.handlers import HANDLER_REASONS, Activity, Reason
from kopf.structs.resources import Resource


def test_on_startup_with_cooldown():
    registry = kopf.get_default_registry()

    with pytest.deprecated_call(match=r"use backoff="):
        @kopf.on.startup(cooldown=78)
        def fn(**_):
            pass

    handlers = registry.activity_handlers.get_handlers(activity=Activity.STARTUP)
    assert len(handlers) == 1
    assert handlers[0].fn is fn
    assert handlers[0].backoff == 78

    with pytest.deprecated_call(match=r"use handler.backoff"):
        assert handlers[0].cooldown == 78


def test_on_cleanup_with_cooldown():
    registry = kopf.get_default_registry()

    with pytest.deprecated_call(match=r"use backoff="):
        @kopf.on.cleanup(cooldown=78)
        def fn(**_):
            pass

    handlers = registry.activity_handlers.get_handlers(activity=Activity.CLEANUP)
    assert len(handlers) == 1
    assert handlers[0].fn is fn
    assert handlers[0].backoff == 78

    with pytest.deprecated_call(match=r"use handler.backoff"):
        assert handlers[0].cooldown == 78


def test_on_probe_with_cooldown():
    registry = kopf.get_default_registry()

    with pytest.deprecated_call(match=r"use backoff="):
        @kopf.on.probe(cooldown=78)
        def fn(**_):
            pass

    handlers = registry.activity_handlers.get_handlers(activity=Activity.PROBE)
    assert len(handlers) == 1
    assert handlers[0].fn is fn
    assert handlers[0].backoff == 78

    with pytest.deprecated_call(match=r"use handler.backoff"):
        assert handlers[0].cooldown == 78


# Resume handlers are mixed-in into all resource-changing reactions with initial listing.
@pytest.mark.parametrize('reason', HANDLER_REASONS)
def test_on_resume_with_cooldown(mocker, reason, cause_factory):
    registry = kopf.get_default_registry()
    resource = Resource('group', 'version', 'plural')
    cause = cause_factory(resource=resource, reason=reason, initial=True)
    mocker.patch('kopf.reactor.registries.match', return_value=True)

    with pytest.deprecated_call(match=r"use backoff="):
        @kopf.on.resume('group', 'version', 'plural', cooldown=78)
        def fn(**_):
            pass

    handlers = registry.resource_changing_handlers[resource].get_handlers(cause)
    assert len(handlers) == 1
    assert handlers[0].fn is fn
    assert handlers[0].backoff == 78

    with pytest.deprecated_call(match=r"use handler.backoff"):
        assert handlers[0].cooldown == 78


def test_on_create_with_cooldown(mocker, cause_factory):
    registry = kopf.get_default_registry()
    resource = Resource('group', 'version', 'plural')
    cause = cause_factory(resource=resource, reason=Reason.CREATE)
    mocker.patch('kopf.reactor.registries.match', return_value=True)

    with pytest.deprecated_call(match=r"use backoff="):
        @kopf.on.create('group', 'version', 'plural', cooldown=78)
        def fn(**_):
            pass

    handlers = registry.resource_changing_handlers[resource].get_handlers(cause)
    assert len(handlers) == 1
    assert handlers[0].fn is fn
    assert handlers[0].backoff == 78

    with pytest.deprecated_call(match=r"use handler.backoff"):
        assert handlers[0].cooldown == 78


def test_on_update_with_cooldown(mocker, cause_factory):
    registry = kopf.get_default_registry()
    resource = Resource('group', 'version', 'plural')
    cause = cause_factory(resource=resource, reason=Reason.UPDATE)
    mocker.patch('kopf.reactor.registries.match', return_value=True)

    with pytest.deprecated_call(match=r"use backoff="):
        @kopf.on.update('group', 'version', 'plural', cooldown=78)
        def fn(**_):
            pass

    handlers = registry.resource_changing_handlers[resource].get_handlers(cause)
    assert len(handlers) == 1
    assert handlers[0].fn is fn
    assert handlers[0].backoff == 78

    with pytest.deprecated_call(match=r"use handler.backoff"):
        assert handlers[0].cooldown == 78


@pytest.mark.parametrize('optional', [
    pytest.param(True, id='optional'),
    pytest.param(False, id='mandatory'),
])
def test_on_delete_with_cooldown(mocker, optional, cause_factory):
    registry = kopf.get_default_registry()
    resource = Resource('group', 'version', 'plural')
    cause = cause_factory(resource=resource, reason=Reason.DELETE)
    mocker.patch('kopf.reactor.registries.match', return_value=True)

    with pytest.deprecated_call(match=r"use backoff="):
        @kopf.on.delete('group', 'version', 'plural', cooldown=78)
        def fn(**_):
            pass

    handlers = registry.resource_changing_handlers[resource].get_handlers(cause)
    assert len(handlers) == 1
    assert handlers[0].fn is fn
    assert handlers[0].backoff == 78

    with pytest.deprecated_call(match=r"use handler.backoff"):
        assert handlers[0].cooldown == 78


def test_on_field_with_cooldown(mocker, cause_factory):
    registry = kopf.get_default_registry()
    resource = Resource('group', 'version', 'plural')
    diff = [('op', ('field', 'subfield'), 'old', 'new')]
    cause = cause_factory(resource=resource, reason=Reason.UPDATE, diff=diff)
    mocker.patch('kopf.reactor.registries.match', return_value=True)

    with pytest.deprecated_call(match=r"use backoff="):
        @kopf.on.field('group', 'version', 'plural', 'field.subfield', cooldown=78)
        def fn(**_):
            pass

    handlers = registry.resource_changing_handlers[resource].get_handlers(cause)
    assert len(handlers) == 1
    assert handlers[0].fn is fn
    assert handlers[0].backoff == 78

    with pytest.deprecated_call(match=r"use handler.backoff"):
        assert handlers[0].cooldown == 78



import asyncio

import freezegun
import pytest

from kopf.engines.peering import Peer, process_peering_event
from kopf.structs import bodies, primitives


@pytest.mark.parametrize('our_name, our_namespace, their_name, their_namespace', [
    ['our-name', 'our-namespace', 'their-name', 'their-namespace'],
    ['our-name', 'our-namespace', 'their-name', 'our-namespace'],
    ['our-name', 'our-namespace', 'their-name', None],
    ['our-name', 'our-namespace', 'our-name', 'their-namespace'],
    ['our-name', 'our-namespace', 'our-name', None],
    ['our-name', None, 'their-name', 'their-namespace'],
    ['our-name', None, 'their-name', 'our-namespace'],
    ['our-name', None, 'their-name', None],
    ['our-name', None, 'our-name', 'their-namespace'],
    ['our-name', None, 'their-name', 'our-namespace'],
])
async def test_other_peering_objects_are_ignored(
        mocker, our_name, our_namespace, their_name, their_namespace):

    ourselves = Peer(id='id', name=our_name, namespace=our_namespace)
    status = mocker.Mock()
    status.items.side_effect = Exception("This should not be called.")
    event = bodies.RawEvent(
        type='ADDED',  # irrelevant
        object={
            'metadata': {'name': their_name, 'namespace': their_namespace},
            'status': status,
        })
    await process_peering_event(
        raw_event=event,
        ourselves=ourselves,
        freeze_mode=primitives.Toggle(),
        replenished=asyncio.Event(),
        autoclean=False,
    )
    assert not status.items.called


@freezegun.freeze_time('2020-12-31T23:59:59.123456')
async def test_toggled_on_for_higher_priority_peer_when_initially_off(caplog, assert_logs):
    event = bodies.RawEvent(
        type='ADDED',  # irrelevant
        object={
            'metadata': {'name': 'name', 'namespace': 'namespace'},  # for matching
            'status': {
                'higher-prio': {
                    'priority': 101,
                    'lifetime': 10,
                    'lastseen': '2020-12-31T23:59:59'
                },
            },
        })

    replenished = asyncio.Event()
    freeze_mode = primitives.Toggle(False)
    ourselves = Peer(id='id', name='name', namespace='namespace', priority=100)

    caplog.set_level(0)
    assert freeze_mode.is_off()
    await process_peering_event(
        raw_event=event,
        freeze_mode=freeze_mode,
        ourselves=ourselves,
        replenished=replenished,
        autoclean=False,
    )
    assert freeze_mode.is_on()
    assert_logs(["Freezing operations in favour of"], prohibited=[
        "Possibly conflicting operators",
        "Freezing all operators, including self",
        "Resuming operations after the freeze",
    ])


@freezegun.freeze_time('2020-12-31T23:59:59.123456')
async def test_ignored_for_higher_priority_peer_when_already_on(caplog, assert_logs):
    event = bodies.RawEvent(
        type='ADDED',  # irrelevant
        object={
            'metadata': {'name': 'name', 'namespace': 'namespace'},  # for matching
            'status': {
                'higher-prio': {
                    'priority': 101,
                    'lifetime': 10,
                    'lastseen': '2020-12-31T23:59:59'
                },
            },
        })

    replenished = asyncio.Event()
    freeze_mode = primitives.Toggle(True)
    ourselves = Peer(id='id', name='name', namespace='namespace', priority=100)

    caplog.set_level(0)
    assert freeze_mode.is_on()
    await process_peering_event(
        raw_event=event,
        freeze_mode=freeze_mode,
        ourselves=ourselves,
        replenished=replenished,
        autoclean=False,
    )
    assert freeze_mode.is_on()
    assert_logs([], prohibited=[
        "Possibly conflicting operators",
        "Freezing all operators, including self",
        "Freezing operations in favour of",
        "Resuming operations after the freeze",
    ])


@freezegun.freeze_time('2020-12-31T23:59:59.123456')
async def test_toggled_off_for_lower_priority_peer_when_initially_on(caplog, assert_logs):
    event = bodies.RawEvent(
        type='ADDED',  # irrelevant
        object={
            'metadata': {'name': 'name', 'namespace': 'namespace'},  # for matching
            'status': {
                'higher-prio': {
                    'priority': 99,
                    'lifetime': 10,
                    'lastseen': '2020-12-31T23:59:59'
                },
            },
        })

    replenished = asyncio.Event()
    freeze_mode = primitives.Toggle(True)
    ourselves = Peer(id='id', name='name', namespace='namespace', priority=100)

    caplog.set_level(0)
    assert freeze_mode.is_on()
    await process_peering_event(
        raw_event=event,
        freeze_mode=freeze_mode,
        ourselves=ourselves,
        replenished=replenished,
        autoclean=False,
    )
    assert freeze_mode.is_off()
    assert_logs(["Resuming operations after the freeze"], prohibited=[
        "Possibly conflicting operators",
        "Freezing all operators, including self",
        "Freezing operations in favour of",
    ])


@freezegun.freeze_time('2020-12-31T23:59:59.123456')
async def test_ignored_for_lower_priority_peer_when_already_off(caplog, assert_logs):
    event = bodies.RawEvent(
        type='ADDED',  # irrelevant
        object={
            'metadata': {'name': 'name', 'namespace': 'namespace'},  # for matching
            'status': {
                'higher-prio': {
                    'priority': 99,
                    'lifetime': 10,
                    'lastseen': '2020-12-31T23:59:59'
                },
            },
        })

    replenished = asyncio.Event()
    freeze_mode = primitives.Toggle(False)
    ourselves = Peer(id='id', name='name', namespace='namespace', priority=100)

    caplog.set_level(0)
    assert freeze_mode.is_off()
    await process_peering_event(
        raw_event=event,
        freeze_mode=freeze_mode,
        ourselves=ourselves,
        replenished=replenished,
        autoclean=False,
    )
    assert freeze_mode.is_off()
    assert_logs([], prohibited=[
        "Possibly conflicting operators",
        "Freezing all operators, including self",
        "Freezing operations in favour of",
        "Resuming operations after the freeze",
    ])


@freezegun.freeze_time('2020-12-31T23:59:59.123456')
async def test_toggled_on_for_same_priority_peer_when_initially_off(caplog, assert_logs):
    event = bodies.RawEvent(
        type='ADDED',  # irrelevant
        object={
            'metadata': {'name': 'name', 'namespace': 'namespace'},  # for matching
            'status': {
                'higher-prio': {
                    'priority': 100,
                    'lifetime': 10,
                    'lastseen': '2020-12-31T23:59:59'
                },
            },
        })

    replenished = asyncio.Event()
    freeze_mode = primitives.Toggle(False)
    ourselves = Peer(id='id', name='name', namespace='namespace', priority=100)

    caplog.set_level(0)
    assert freeze_mode.is_off()
    await process_peering_event(
        raw_event=event,
        freeze_mode=freeze_mode,
        ourselves=ourselves,
        replenished=replenished,
        autoclean=False,
    )
    assert freeze_mode.is_on()
    assert_logs([
        "Possibly conflicting operators",
        "Freezing all operators, including self",
    ], prohibited=[
        "Freezing operations in favour of",
        "Resuming operations after the freeze",
    ])


@freezegun.freeze_time('2020-12-31T23:59:59.123456')
async def test_ignored_for_same_priority_peer_when_already_on(caplog, assert_logs):
    event = bodies.RawEvent(
        type='ADDED',  # irrelevant
        object={
            'metadata': {'name': 'name', 'namespace': 'namespace'},  # for matching
            'status': {
                'higher-prio': {
                    'priority': 100,
                    'lifetime': 10,
                    'lastseen': '2020-12-31T23:59:59'
                },
            },
        })

    replenished = asyncio.Event()
    freeze_mode = primitives.Toggle(True)
    ourselves = Peer(id='id', name='name', namespace='namespace', priority=100)

    caplog.set_level(0)
    assert freeze_mode.is_on()
    await process_peering_event(
        raw_event=event,
        freeze_mode=freeze_mode,
        ourselves=ourselves,
        replenished=replenished,
        autoclean=False,
    )
    assert freeze_mode.is_on()
    assert_logs([
        "Possibly conflicting operators",
    ], prohibited=[
        "Freezing all operators, including self",
        "Freezing operations in favour of",
        "Resuming operations after the freeze",
    ])



import urllib.parse
from typing import List, Mapping, NamedTuple, Optional


# An immutable reference to a custom resource definition.
class Resource(NamedTuple):
    group: str
    version: str
    plural: str

    @property
    def name(self) -> str:
        return f'{self.plural}.{self.group}'.strip('.')

    @property
    def api_version(self) -> str:
        # Strip heading/trailing slashes if group is absent (e.g. for pods).
        return f'{self.group}/{self.version}'.strip('/')

    def get_url(
            self,
            *,
            server: Optional[str] = None,
            namespace: Optional[str] = None,
            name: Optional[str] = None,
            subresource: Optional[str] = None,
            params: Optional[Mapping[str, str]] = None,
    ) -> str:
        if subresource is not None and name is None:
            raise ValueError("Subresources can be used only with specific resources by their name.")

        return self._build_url(server, params, [
            '/api' if self.group == '' and self.version == 'v1' else '/apis',
            self.group,
            self.version,
            'namespaces' if namespace is not None else None,
            namespace,
            self.plural,
            name,
            subresource,
        ])

    def get_version_url(
            self,
            *,
            server: Optional[str] = None,
            params: Optional[Mapping[str, str]] = None,
    ) -> str:
        return self._build_url(server, params, [
            '/api' if self.group == '' and self.version == 'v1' else '/apis',
            self.group,
            self.version,
        ])

    def _build_url(
            self,
            server: Optional[str],
            params: Optional[Mapping[str, str]],
            parts: List[Optional[str]],
    ) -> str:
        query = urllib.parse.urlencode(params, encoding='utf-8') if params else ''
        path = '/'.join([part for part in parts if part])
        url = path + ('?' if query else '') + query
        return url if server is None else server.rstrip('/') + '/' + url.lstrip('/')



import pytest

from kopf.engines.peering import Peer, peers_keepalive


class StopInfiniteCycleException(Exception):
    pass


async def test_background_task_runs(mocker):
    ourselves = Peer(id='id', name='name', namespace='namespace', lifetime=33)
    keepalive_mock = mocker.patch.object(ourselves, 'keepalive')
    disappear_mock = mocker.patch.object(ourselves, 'disappear')

    sleep_mock = mocker.patch('asyncio.sleep')
    sleep_mock.side_effect = [None, None, StopInfiniteCycleException]

    with pytest.raises(StopInfiniteCycleException):
        await peers_keepalive(ourselves=ourselves)

    assert sleep_mock.call_count == 3
    assert sleep_mock.call_args_list[0][0][0] == 33 - 10
    assert sleep_mock.call_args_list[1][0][0] == 33 - 10
    assert sleep_mock.call_args_list[2][0][0] == 33 - 10

    assert keepalive_mock.call_count == 3
    assert disappear_mock.call_count == 1



"""
Only the tests from the K8s client library (simulated) to the event stream.

Excluded: the watch-stream consumption and queue multiplexing routines
(see ``tests_queueing.py``).

Used for internal control that the event multiplexing works are intended.
If the intentions change, the tests should be rewritten.
They are NOT part of the public interface of the framework.
"""
import asyncio
import logging

import aiohttp
import pytest

from kopf.clients.watching import WatchingError, infinite_watch, streaming_watch
from kopf.structs.primitives import Toggle

STREAM_WITH_NORMAL_EVENTS = [
    {'type': 'ADDED', 'object': {'spec': 'a'}},
    {'type': 'ADDED', 'object': {'spec': 'b'}},
]
STREAM_WITH_UNKNOWN_EVENT = [
    {'type': 'ADDED', 'object': {'spec': 'a'}},
    {'type': 'UNKNOWN', 'object': {}},
    {'type': 'ADDED', 'object': {'spec': 'b'}},
]
STREAM_WITH_ERROR_410GONE = [
    {'type': 'ADDED', 'object': {'spec': 'a'}},
    {'type': 'ERROR', 'object': {'code': 410}},
    {'type': 'ADDED', 'object': {'spec': 'b'}},
]
STREAM_WITH_ERROR_CODE = [
    {'type': 'ADDED', 'object': {'spec': 'a'}},
    {'type': 'ERROR', 'object': {'code': 666}},
    {'type': 'ADDED', 'object': {'spec': 'b'}},
]


class SampleException(Exception):
    pass


@pytest.fixture(params=[
    pytest.param('something', id='namespace'),
    pytest.param(None, id='cluster'),
])
def namespace(request):
    return request.param


async def test_empty_stream_yields_nothing(settings, resource, stream, namespace):
    
    stream.feed([], namespace=namespace)
    stream.close(namespace=namespace)

    events = []
    async for event in streaming_watch(settings=settings,
                                       resource=resource,
                                       namespace=namespace):
        events.append(event)

    assert len(events) == 0


async def test_event_stream_yields_everything(
        settings, resource, stream, namespace):

    stream.feed(STREAM_WITH_NORMAL_EVENTS, namespace=namespace)
    stream.close(namespace=namespace)

    events = []
    async for event in streaming_watch(settings=settings,
                                       resource=resource,
                                       namespace=namespace):
        events.append(event)

    assert len(events) == 2
    assert events[0]['object']['spec'] == 'a'
    assert events[1]['object']['spec'] == 'b'


async def test_unknown_event_type_ignored(
        settings, resource, stream, namespace, caplog):

    caplog.set_level(logging.DEBUG)
    stream.feed(STREAM_WITH_UNKNOWN_EVENT, namespace=namespace)
    stream.close(namespace=namespace)

    events = []
    async for event in streaming_watch(settings=settings,
                                       resource=resource,
                                       namespace=namespace):
        events.append(event)

    assert len(events) == 2
    assert events[0]['object']['spec'] == 'a'
    assert events[1]['object']['spec'] == 'b'
    assert "Ignoring an unsupported event type" in caplog.text
    assert "UNKNOWN" in caplog.text


async def test_error_410gone_exits_normally(
        settings, resource, stream, namespace, caplog):

    caplog.set_level(logging.DEBUG)
    stream.feed(STREAM_WITH_ERROR_410GONE, namespace=namespace)
    stream.close(namespace=namespace)

    events = []
    async for event in streaming_watch(settings=settings,
                                       resource=resource,
                                       namespace=namespace):
        events.append(event)

    assert len(events) == 1
    assert events[0]['object']['spec'] == 'a'
    assert "Restarting the watch-stream" in caplog.text


async def test_unknown_error_raises_exception(
        settings, resource, stream, namespace):

    stream.feed(STREAM_WITH_ERROR_CODE, namespace=namespace)
    stream.close(namespace=namespace)

    events = []
    with pytest.raises(WatchingError) as e:
        async for event in streaming_watch(settings=settings,
                                           resource=resource,
                                           namespace=namespace):
            events.append(event)

    assert len(events) == 1
    assert events[0]['object']['spec'] == 'a'
    assert '666' in str(e.value)


async def test_exception_escalates(
        settings, resource, stream, namespace, enforced_session, mocker):

    enforced_session.get = mocker.Mock(side_effect=SampleException())
    stream.feed([], namespace=namespace)
    stream.close(namespace=namespace)

    events = []
    with pytest.raises(SampleException):
        async for event in streaming_watch(settings=settings,
                                           resource=resource,
                                           namespace=namespace):
            events.append(event)

    assert len(events) == 0


async def test_freezing_is_ignored_if_turned_off(
        settings, resource, stream, namespace, timer, caplog, assert_logs):

    stream.feed(STREAM_WITH_NORMAL_EVENTS, namespace=namespace)
    stream.close(namespace=namespace)

    freeze_mode = Toggle(False)
    events = []

    async def read_stream():
        async for event in streaming_watch(settings=settings,
                                           resource=resource,
                                           namespace=namespace,
                                           freeze_mode=freeze_mode):
            events.append(event)

    caplog.set_level(logging.DEBUG)
    with timer:
        await asyncio.wait_for(read_stream(), timeout=0.5)

    assert len(events) == 2
    assert timer.seconds < 0.2  # no waits, exits as soon as possible
    assert_logs([], prohibited=[
        r"Freezing the watch-stream for",
        r"Resuming the watch-stream for",
    ])


async def test_freezing_waits_forever_if_not_resumed(
        settings, resource, stream, namespace, timer, caplog, assert_logs):

    stream.feed(STREAM_WITH_NORMAL_EVENTS, namespace=namespace)
    stream.close(namespace=namespace)

    freeze_mode = Toggle(True)
    events = []

    async def read_stream():
        async for event in streaming_watch(settings=settings,
                                           resource=resource,
                                           namespace=namespace,
                                           freeze_mode=freeze_mode):
            events.append(event)

    caplog.set_level(logging.DEBUG)
    with timer:
        with pytest.raises(asyncio.TimeoutError):
            await asyncio.wait_for(read_stream(), timeout=0.5)

    assert len(events) == 0
    assert timer.seconds >= 0.5
    assert_logs([
        r"Freezing the watch-stream for",
    ], prohibited=[
        r"Resuming the watch-stream for",
    ])


async def test_freezing_waits_until_resumed(
        settings, resource, stream, namespace, timer, caplog, assert_logs):

    stream.feed(STREAM_WITH_NORMAL_EVENTS, namespace=namespace)
    stream.close(namespace=namespace)

    freeze_mode = Toggle(True)
    events = []

    async def delayed_resuming(delay: float):
        await asyncio.sleep(delay)
        await freeze_mode.turn_off()

    async def read_stream():
        async for event in streaming_watch(settings=settings,
                                           resource=resource,
                                           namespace=namespace,
                                           freeze_mode=freeze_mode):
            events.append(event)

    caplog.set_level(logging.DEBUG)
    with timer:
        asyncio.create_task(delayed_resuming(0.2))
        await asyncio.wait_for(read_stream(), timeout=1.0)

    assert len(events) == 2
    assert timer.seconds >= 0.2
    assert timer.seconds <= 0.5
    assert_logs([
        r"Freezing the watch-stream for",
        r"Resuming the watch-stream for",
    ])


async def test_infinite_watch_never_exits_normally(
        settings, resource, stream, namespace, aresponses):

    error = aresponses.Response(status=555, reason='stop-infinite-cycle')
    stream.feed(
        STREAM_WITH_ERROR_410GONE,          # watching restarted
        STREAM_WITH_UNKNOWN_EVENT,          # event ignored
        error,                              # to finally exit it somehow
        namespace=namespace,
    )
    stream.close(namespace=namespace)

    events = []
    with pytest.raises(aiohttp.ClientResponseError) as e:
        async for event in infinite_watch(settings=settings,
                                          resource=resource,
                                          namespace=namespace):
            events.append(event)

    assert e.value.status == 555
    assert e.value.message == 'stop-infinite-cycle'

    assert len(events) == 3
    assert events[0]['object']['spec'] == 'a'
    assert events[1]['object']['spec'] == 'a'
    assert events[2]['object']['spec'] == 'b'


# See: See: https://github.com/zalando-incubator/kopf/issues/275
async def test_long_line_parsing(
        settings, resource, stream, namespace, aresponses):

    content = [
        {'type': 'ADDED', 'object': {'spec': {'field': 'x'}}},
        {'type': 'ADDED', 'object': {'spec': {'field': 'y' * (2 * 1024 * 1024)}}},
        {'type': 'ADDED', 'object': {'spec': {'field': 'z' * (4 * 1024 * 1024)}}},
    ]
    stream.feed(content, namespace=namespace)
    stream.close(namespace=namespace)

    events = []
    async for event in streaming_watch(settings=settings,
                                       resource=resource,
                                       namespace=namespace):
        events.append(event)

    assert len(events) == 3
    assert len(events[0]['object']['spec']['field']) == 1
    assert len(events[1]['object']['spec']['field']) == 2 * 1024 * 1024
    assert len(events[2]['object']['spec']['field']) == 4 * 1024 * 1024



"""
The decorators for the event handlers. Usually used as::

    import kopf

    @kopf.on.create('zalando.org', 'v1', 'kopfexamples')
    def creation_handler(**kwargs):
        pass

This module is a part of the framework's public interface.
"""

# TODO: add cluster=True support (different API methods)
import inspect
import warnings
from typing import Any, Callable, Optional

from kopf.reactor import handling, registries
from kopf.structs import callbacks, dicts, filters, handlers, resources

ActivityDecorator = Callable[[callbacks.ActivityFn], callbacks.ActivityFn]
ResourceWatchingDecorator = Callable[[callbacks.ResourceWatchingFn], callbacks.ResourceWatchingFn]
ResourceChangingDecorator = Callable[[callbacks.ResourceChangingFn], callbacks.ResourceChangingFn]
ResourceDaemonDecorator = Callable[[callbacks.ResourceDaemonFn], callbacks.ResourceDaemonFn]
ResourceTimerDecorator = Callable[[callbacks.ResourceTimerFn], callbacks.ResourceTimerFn]


def startup(  # lgtm[py/similar-function]
        *,
        id: Optional[str] = None,
        errors: Optional[handlers.ErrorsMode] = None,
        timeout: Optional[float] = None,
        retries: Optional[int] = None,
        backoff: Optional[float] = None,
        cooldown: Optional[float] = None,  # deprecated, use `backoff`
        registry: Optional[registries.OperatorRegistry] = None,
) -> ActivityDecorator:
    def decorator(  # lgtm[py/similar-function]
            fn: callbacks.ActivityFn,
    ) -> callbacks.ActivityFn:
        _warn_deprecated_signatures(fn)
        real_registry = registry if registry is not None else registries.get_default_registry()
        real_id = registries.generate_id(fn=fn, id=id)
        handler = handlers.ActivityHandler(
            fn=fn, id=real_id,
            errors=errors, timeout=timeout, retries=retries, backoff=backoff, cooldown=cooldown,
            activity=handlers.Activity.STARTUP,
        )
        real_registry.activity_handlers.append(handler)
        return fn
    return decorator


def cleanup(  # lgtm[py/similar-function]
        *,
        id: Optional[str] = None,
        errors: Optional[handlers.ErrorsMode] = None,
        timeout: Optional[float] = None,
        retries: Optional[int] = None,
        backoff: Optional[float] = None,
        cooldown: Optional[float] = None,  # deprecated, use `backoff`
        registry: Optional[registries.OperatorRegistry] = None,
) -> ActivityDecorator:
    def decorator(  # lgtm[py/similar-function]
            fn: callbacks.ActivityFn,
    ) -> callbacks.ActivityFn:
        _warn_deprecated_signatures(fn)
        real_registry = registry if registry is not None else registries.get_default_registry()
        real_id = registries.generate_id(fn=fn, id=id)
        handler = handlers.ActivityHandler(
            fn=fn, id=real_id,
            errors=errors, timeout=timeout, retries=retries, backoff=backoff, cooldown=cooldown,
            activity=handlers.Activity.CLEANUP,
        )
        real_registry.activity_handlers.append(handler)
        return fn
    return decorator


def login(  # lgtm[py/similar-function]
        *,
        id: Optional[str] = None,
        errors: Optional[handlers.ErrorsMode] = None,
        timeout: Optional[float] = None,
        retries: Optional[int] = None,
        backoff: Optional[float] = None,
        cooldown: Optional[float] = None,  # deprecated, use `backoff`
        registry: Optional[registries.OperatorRegistry] = None,
) -> ActivityDecorator:
    """ ``@kopf.on.login()`` handler for custom (re-)authentication. """
    def decorator(  # lgtm[py/similar-function]
            fn: callbacks.ActivityFn,
    ) -> callbacks.ActivityFn:
        _warn_deprecated_signatures(fn)
        real_registry = registry if registry is not None else registries.get_default_registry()
        real_id = registries.generate_id(fn=fn, id=id)
        handler = handlers.ActivityHandler(
            fn=fn, id=real_id,
            errors=errors, timeout=timeout, retries=retries, backoff=backoff, cooldown=cooldown,
            activity=handlers.Activity.AUTHENTICATION,
        )
        real_registry.activity_handlers.append(handler)
        return fn
    return decorator


def probe(  # lgtm[py/similar-function]
        *,
        id: Optional[str] = None,
        errors: Optional[handlers.ErrorsMode] = None,
        timeout: Optional[float] = None,
        retries: Optional[int] = None,
        backoff: Optional[float] = None,
        cooldown: Optional[float] = None,  # deprecated, use `backoff`
        registry: Optional[registries.OperatorRegistry] = None,
) -> ActivityDecorator:
    """ ``@kopf.on.probe()`` handler for arbitrary liveness metrics. """
    def decorator(  # lgtm[py/similar-function]
            fn: callbacks.ActivityFn,
    ) -> callbacks.ActivityFn:
        _warn_deprecated_signatures(fn)
        real_registry = registry if registry is not None else registries.get_default_registry()
        real_id = registries.generate_id(fn=fn, id=id)
        handler = handlers.ActivityHandler(
            fn=fn, id=real_id,
            errors=errors, timeout=timeout, retries=retries, backoff=backoff, cooldown=cooldown,
            activity=handlers.Activity.PROBE,
        )
        real_registry.activity_handlers.append(handler)
        return fn
    return decorator


def resume(  # lgtm[py/similar-function]
        group: str, version: str, plural: str,
        *,
        id: Optional[str] = None,
        errors: Optional[handlers.ErrorsMode] = None,
        timeout: Optional[float] = None,
        retries: Optional[int] = None,
        backoff: Optional[float] = None,
        cooldown: Optional[float] = None,  # deprecated, use `backoff`
        registry: Optional[registries.OperatorRegistry] = None,
        deleted: Optional[bool] = None,
        labels: Optional[filters.MetaFilter] = None,
        annotations: Optional[filters.MetaFilter] = None,
        when: Optional[callbacks.WhenFilterFn] = None,
) -> ResourceChangingDecorator:
    """ ``@kopf.on.resume()`` handler for the object resuming on operator (re)start. """
    def decorator(  # lgtm[py/similar-function]
            fn: callbacks.ResourceChangingFn,
    ) -> callbacks.ResourceChangingFn:
        _warn_deprecated_signatures(fn)
        _warn_deprecated_filters(labels, annotations)
        real_registry = registry if registry is not None else registries.get_default_registry()
        real_resource = resources.Resource(group, version, plural)
        real_id = registries.generate_id(fn=fn, id=id)
        handler = handlers.ResourceChangingHandler(
            fn=fn, id=real_id, field=None,
            errors=errors, timeout=timeout, retries=retries, backoff=backoff, cooldown=cooldown,
            labels=labels, annotations=annotations, when=when,
            initial=True, deleted=deleted, requires_finalizer=None,
            reason=None,
        )
        real_registry.resource_changing_handlers[real_resource].append(handler)
        return fn
    return decorator


def create(  # lgtm[py/similar-function]
        group: str, version: str, plural: str,
        *,
        id: Optional[str] = None,
        errors: Optional[handlers.ErrorsMode] = None,
        timeout: Optional[float] = None,
        retries: Optional[int] = None,
        backoff: Optional[float] = None,
        cooldown: Optional[float] = None,  # deprecated; use backoff.
        registry: Optional[registries.OperatorRegistry] = None,
        labels: Optional[filters.MetaFilter] = None,
        annotations: Optional[filters.MetaFilter] = None,
        when: Optional[callbacks.WhenFilterFn] = None,
) -> ResourceChangingDecorator:
    """ ``@kopf.on.create()`` handler for the object creation. """
    def decorator(  # lgtm[py/similar-function]
            fn: callbacks.ResourceChangingFn,
    ) -> callbacks.ResourceChangingFn:
        _warn_deprecated_signatures(fn)
        _warn_deprecated_filters(labels, annotations)
        real_registry = registry if registry is not None else registries.get_default_registry()
        real_resource = resources.Resource(group, version, plural)
        real_id = registries.generate_id(fn=fn, id=id)
        handler = handlers.ResourceChangingHandler(
            fn=fn, id=real_id, field=None,
            errors=errors, timeout=timeout, retries=retries, backoff=backoff, cooldown=cooldown,
            labels=labels, annotations=annotations, when=when,
            initial=None, deleted=None, requires_finalizer=None,
            reason=handlers.Reason.CREATE,
        )
        real_registry.resource_changing_handlers[real_resource].append(handler)
        return fn
    return decorator


def update(  # lgtm[py/similar-function]
        group: str, version: str, plural: str,
        *,
        id: Optional[str] = None,
        errors: Optional[handlers.ErrorsMode] = None,
        timeout: Optional[float] = None,
        retries: Optional[int] = None,
        backoff: Optional[float] = None,
        cooldown: Optional[float] = None,  # deprecated, use `backoff`
        registry: Optional[registries.OperatorRegistry] = None,
        labels: Optional[filters.MetaFilter] = None,
        annotations: Optional[filters.MetaFilter] = None,
        when: Optional[callbacks.WhenFilterFn] = None,
) -> ResourceChangingDecorator:
    """ ``@kopf.on.update()`` handler for the object update or change. """
    def decorator(  # lgtm[py/similar-function]
            fn: callbacks.ResourceChangingFn,
    ) -> callbacks.ResourceChangingFn:
        _warn_deprecated_signatures(fn)
        _warn_deprecated_filters(labels, annotations)
        real_registry = registry if registry is not None else registries.get_default_registry()
        real_resource = resources.Resource(group, version, plural)
        real_id = registries.generate_id(fn=fn, id=id)
        handler = handlers.ResourceChangingHandler(
            fn=fn, id=real_id, field=None,
            errors=errors, timeout=timeout, retries=retries, backoff=backoff, cooldown=cooldown,
            labels=labels, annotations=annotations, when=when,
            initial=None, deleted=None, requires_finalizer=None,
            reason=handlers.Reason.UPDATE,
        )
        real_registry.resource_changing_handlers[real_resource].append(handler)
        return fn
    return decorator


def delete(  # lgtm[py/similar-function]
        group: str, version: str, plural: str,
        *,
        id: Optional[str] = None,
        errors: Optional[handlers.ErrorsMode] = None,
        timeout: Optional[float] = None,
        retries: Optional[int] = None,
        backoff: Optional[float] = None,
        cooldown: Optional[float] = None,  # deprecated, use `backoff`
        registry: Optional[registries.OperatorRegistry] = None,
        optional: Optional[bool] = None,
        labels: Optional[filters.MetaFilter] = None,
        annotations: Optional[filters.MetaFilter] = None,
        when: Optional[callbacks.WhenFilterFn] = None,
) -> ResourceChangingDecorator:
    """ ``@kopf.on.delete()`` handler for the object deletion. """
    def decorator(  # lgtm[py/similar-function]
            fn: callbacks.ResourceChangingFn,
    ) -> callbacks.ResourceChangingFn:
        _warn_deprecated_signatures(fn)
        _warn_deprecated_filters(labels, annotations)
        real_registry = registry if registry is not None else registries.get_default_registry()
        real_resource = resources.Resource(group, version, plural)
        real_id = registries.generate_id(fn=fn, id=id)
        handler = handlers.ResourceChangingHandler(
            fn=fn, id=real_id, field=None,
            errors=errors, timeout=timeout, retries=retries, backoff=backoff, cooldown=cooldown,
            labels=labels, annotations=annotations, when=when,
            initial=None, deleted=None, requires_finalizer=bool(not optional),
            reason=handlers.Reason.DELETE,
        )
        real_registry.resource_changing_handlers[real_resource].append(handler)
        return fn
    return decorator


def field(  # lgtm[py/similar-function]
        group: str, version: str, plural: str,
        field: dicts.FieldSpec,
        *,
        id: Optional[str] = None,
        errors: Optional[handlers.ErrorsMode] = None,
        timeout: Optional[float] = None,
        retries: Optional[int] = None,
        backoff: Optional[float] = None,
        cooldown: Optional[float] = None,  # deprecated, use `backoff`
        registry: Optional[registries.OperatorRegistry] = None,
        labels: Optional[filters.MetaFilter] = None,
        annotations: Optional[filters.MetaFilter] = None,
        when: Optional[callbacks.WhenFilterFn] = None,
) -> ResourceChangingDecorator:
    """ ``@kopf.on.field()`` handler for the individual field changes. """
    def decorator(  # lgtm[py/similar-function]
            fn: callbacks.ResourceChangingFn,
    ) -> callbacks.ResourceChangingFn:
        _warn_deprecated_signatures(fn)
        _warn_deprecated_filters(labels, annotations)
        real_registry = registry if registry is not None else registries.get_default_registry()
        real_resource = resources.Resource(group, version, plural)
        real_field = dicts.parse_field(field) or None  # to not store tuple() as a no-field case.
        real_id = registries.generate_id(fn=fn, id=id, suffix=".".join(real_field or []))
        handler = handlers.ResourceChangingHandler(
            fn=fn, id=real_id, field=real_field,
            errors=errors, timeout=timeout, retries=retries, backoff=backoff, cooldown=cooldown,
            labels=labels, annotations=annotations, when=when,
            initial=None, deleted=None, requires_finalizer=None,
            reason=None,
        )
        real_registry.resource_changing_handlers[real_resource].append(handler)
        return fn
    return decorator


def event(  # lgtm[py/similar-function]
        group: str, version: str, plural: str,
        *,
        id: Optional[str] = None,
        registry: Optional[registries.OperatorRegistry] = None,
        labels: Optional[filters.MetaFilter] = None,
        annotations: Optional[filters.MetaFilter] = None,
        when: Optional[callbacks.WhenFilterFn] = None,
) -> ResourceWatchingDecorator:
    """ ``@kopf.on.event()`` handler for the silent spies on the events. """
    def decorator(  # lgtm[py/similar-function]
            fn: callbacks.ResourceWatchingFn,
    ) -> callbacks.ResourceWatchingFn:
        _warn_deprecated_signatures(fn)
        _warn_deprecated_filters(labels, annotations)
        real_registry = registry if registry is not None else registries.get_default_registry()
        real_resource = resources.Resource(group, version, plural)
        real_id = registries.generate_id(fn=fn, id=id)
        handler = handlers.ResourceWatchingHandler(
            fn=fn, id=real_id,
            errors=None, timeout=None, retries=None, backoff=None, cooldown=None,
            labels=labels, annotations=annotations, when=when,
        )
        real_registry.resource_watching_handlers[real_resource].append(handler)
        return fn
    return decorator


def daemon(  # lgtm[py/similar-function]
        group: str, version: str, plural: str,
        *,
        id: Optional[str] = None,
        errors: Optional[handlers.ErrorsMode] = None,
        timeout: Optional[float] = None,
        retries: Optional[int] = None,
        backoff: Optional[float] = None,
        cooldown: Optional[float] = None,  # deprecated, use `backoff`
        registry: Optional[registries.OperatorRegistry] = None,
        labels: Optional[filters.MetaFilter] = None,
        annotations: Optional[filters.MetaFilter] = None,
        when: Optional[callbacks.WhenFilterFn] = None,
        initial_delay: Optional[float] = None,
        cancellation_backoff: Optional[float] = None,
        cancellation_timeout: Optional[float] = None,
        cancellation_polling: Optional[float] = None,
) -> ResourceDaemonDecorator:
    """ ``@kopf.daemon()`` decorator for the background threads/tasks. """
    def decorator(  # lgtm[py/similar-function]
            fn: callbacks.ResourceDaemonFn,
    ) -> callbacks.ResourceDaemonFn:
        _warn_deprecated_signatures(fn)
        _warn_deprecated_filters(labels, annotations)
        real_registry = registry if registry is not None else registries.get_default_registry()
        real_resource = resources.Resource(group, version, plural)
        real_id = registries.generate_id(fn=fn, id=id)
        handler = handlers.ResourceDaemonHandler(
            fn=fn, id=real_id,
            errors=errors, timeout=timeout, retries=retries, backoff=backoff, cooldown=cooldown,
            labels=labels, annotations=annotations, when=when,
            initial_delay=initial_delay, requires_finalizer=True,
            cancellation_backoff=cancellation_backoff,
            cancellation_timeout=cancellation_timeout,
            cancellation_polling=cancellation_polling,
        )
        real_registry.resource_spawning_handlers[real_resource].append(handler)
        return fn
    return decorator


def timer(  # lgtm[py/similar-function]
        group: str, version: str, plural: str,
        *,
        id: Optional[str] = None,
        errors: Optional[handlers.ErrorsMode] = None,
        timeout: Optional[float] = None,
        retries: Optional[int] = None,
        backoff: Optional[float] = None,
        cooldown: Optional[float] = None,  # deprecated, use `backoff`
        registry: Optional[registries.OperatorRegistry] = None,
        labels: Optional[filters.MetaFilter] = None,
        annotations: Optional[filters.MetaFilter] = None,
        when: Optional[callbacks.WhenFilterFn] = None,
        initial_delay: Optional[float] = None,
        sharp: Optional[bool] = None,
        idle: Optional[float] = None,
        interval: Optional[float] = None,
) -> ResourceTimerDecorator:
    """ ``@kopf.timer()`` handler for the regular events. """
    def decorator(  # lgtm[py/similar-function]
            fn: callbacks.ResourceTimerFn,
    ) -> callbacks.ResourceTimerFn:
        _warn_deprecated_signatures(fn)
        _warn_deprecated_filters(labels, annotations)
        real_registry = registry if registry is not None else registries.get_default_registry()
        real_resource = resources.Resource(group, version, plural)
        real_id = registries.generate_id(fn=fn, id=id)
        handler = handlers.ResourceTimerHandler(
            fn=fn, id=real_id,
            errors=errors, timeout=timeout, retries=retries, backoff=backoff, cooldown=cooldown,
            labels=labels, annotations=annotations, when=when,
            initial_delay=initial_delay, requires_finalizer=True,
            sharp=sharp, idle=idle, interval=interval,
        )
        real_registry.resource_spawning_handlers[real_resource].append(handler)
        return fn
    return decorator


# TODO: find a better name: `@kopf.on.this` is confusing and does not fully
# TODO: match with the `@kopf.on.{cause}` pattern, where cause is create/update/delete.
def this(  # lgtm[py/similar-function]
        *,
        id: Optional[str] = None,
        errors: Optional[handlers.ErrorsMode] = None,
        timeout: Optional[float] = None,
        retries: Optional[int] = None,
        backoff: Optional[float] = None,
        cooldown: Optional[float] = None,  # deprecated, use `backoff`
        registry: Optional[registries.ResourceChangingRegistry] = None,
        labels: Optional[filters.MetaFilter] = None,
        annotations: Optional[filters.MetaFilter] = None,
        when: Optional[callbacks.WhenFilterFn] = None,
) -> ResourceChangingDecorator:
    """
    ``@kopf.on.this()`` decorator for the dynamically generated sub-handlers.

    Can be used only inside of the handler function.
    It is efficiently a syntax sugar to look like all other handlers::

        @kopf.on.create('zalando.org', 'v1', 'kopfexamples')
        def create(*, spec, **kwargs):

            for task in spec.get('tasks', []):

                @kopf.on.this(id=f'task_{task}')
                def create_task(*, spec, task=task, **kwargs):

                    pass

    In this example, having spec.tasks set to ``[abc, def]``, this will create
    the following handlers: ``create``, ``create/task_abc``, ``create/task_def``.

    The parent handler is not considered as finished if there are unfinished
    sub-handlers left. Since the sub-handlers will be executed in the regular
    reactor and lifecycle, with multiple low-level events (one per iteration),
    the parent handler will also be executed multiple times, and is expected
    to produce the same (or at least predictable) set of sub-handlers.
    In addition, keep its logic idempotent (not failing on the repeated calls).

    Note: ``task=task`` is needed to freeze the closure variable, so that every
    create function will have its own value, not the latest in the for-cycle.
    """
    def decorator(  # lgtm[py/similar-function]
            fn: callbacks.ResourceChangingFn,
    ) -> callbacks.ResourceChangingFn:
        _warn_deprecated_signatures(fn)
        _warn_deprecated_filters(labels, annotations)
        parent_handler = handling.handler_var.get()
        real_registry = registry if registry is not None else handling.subregistry_var.get()
        real_id = registries.generate_id(fn=fn, id=id,
                                         prefix=parent_handler.id if parent_handler else None)
        handler = handlers.ResourceChangingHandler(
            fn=fn, id=real_id, field=None,
            errors=errors, timeout=timeout, retries=retries, backoff=backoff, cooldown=cooldown,
            labels=labels, annotations=annotations, when=when,
            initial=None, deleted=None, requires_finalizer=None,
            reason=None,
        )
        real_registry.append(handler)
        return fn
    return decorator


def register(  # lgtm[py/similar-function]
        fn: callbacks.ResourceChangingFn,
        *,
        id: Optional[str] = None,
        errors: Optional[handlers.ErrorsMode] = None,
        timeout: Optional[float] = None,
        retries: Optional[int] = None,
        backoff: Optional[float] = None,
        cooldown: Optional[float] = None,  # deprecated, use `backoff`
        registry: Optional[registries.ResourceChangingRegistry] = None,
        labels: Optional[filters.MetaFilter] = None,
        annotations: Optional[filters.MetaFilter] = None,
        when: Optional[callbacks.WhenFilterFn] = None,
) -> callbacks.ResourceChangingFn:
    """
    Register a function as a sub-handler of the currently executed handler.

    Example::

        @kopf.on.create('zalando.org', 'v1', 'kopfexamples')
        def create_it(spec, **kwargs):
            for task in spec.get('tasks', []):

                def create_single_task(task=task, **_):
                    pass

                kopf.register(id=task, fn=create_single_task)

    This is efficiently an equivalent for::

        @kopf.on.create('zalando.org', 'v1', 'kopfexamples')
        def create_it(spec, **kwargs):
            for task in spec.get('tasks', []):

                @kopf.on.this(id=task)
                def create_single_task(task=task, **_):
                    pass
    """
    decorator = this(
        id=id, registry=registry,
        errors=errors, timeout=timeout, retries=retries, backoff=backoff, cooldown=cooldown,
        labels=labels, annotations=annotations, when=when,
    )
    return decorator(fn)


def _warn_deprecated_signatures(
        fn: Callable[..., Any],
) -> None:
    argspec = inspect.getfullargspec(fn)
    if 'cause' in argspec.args or 'cause' in argspec.kwonlyargs:
        warnings.warn("`cause` kwarg is deprecated; use kwargs directly.", DeprecationWarning)


def _warn_deprecated_filters(
        labels: Optional[filters.MetaFilter],
        annotations: Optional[filters.MetaFilter],
) -> None:
    if labels is not None:
        for key, val in labels.items():
            if val is None:
                warnings.warn(
                    f"`None` for label filters is deprecated; use kopf.PRESENT.",
                    DeprecationWarning, stacklevel=2)
    if annotations is not None:
        for key, val in annotations.items():
            if val is None:
                warnings.warn(
                    f"`None` for annotation filters is deprecated; use kopf.PRESENT.",
                    DeprecationWarning, stacklevel=2)



import pytest

from kopf.engines.peering import CLUSTER_PEERING_RESOURCE, NAMESPACED_PEERING_RESOURCE


@pytest.fixture(autouse=True)
def _autouse_fake_vault(fake_vault):
    pass


@pytest.fixture()
def with_cluster_crd(hostname, aresponses):
    result = {'resources': [{
        'name': CLUSTER_PEERING_RESOURCE.plural,
        'namespaced': False,
    }]}
    url = CLUSTER_PEERING_RESOURCE.get_version_url()
    aresponses.add(hostname, url, 'get', result)


@pytest.fixture()
def with_namespaced_crd(hostname, aresponses):
    result = {'resources': [{
        'name': NAMESPACED_PEERING_RESOURCE.plural,
        'namespaced': True,
    }]}
    url = NAMESPACED_PEERING_RESOURCE.get_version_url()
    aresponses.add(hostname, url, 'get', result)


@pytest.fixture()
def with_both_crds(hostname, aresponses):
    result = {'resources': [{
        'name': CLUSTER_PEERING_RESOURCE.plural,
        'namespaced': False,
    }, {
        'name': NAMESPACED_PEERING_RESOURCE.plural,
        'namespaced': True,
    }]}
    urls = {
        CLUSTER_PEERING_RESOURCE.get_version_url(),
        NAMESPACED_PEERING_RESOURCE.get_version_url(),
    }
    for url in urls:
        aresponses.add(hostname, url, 'get', result)



import asyncio

import pytest

from kopf.structs.primitives import Toggle


async def test_created_as_off():
    toggle = Toggle()

    assert not toggle
    assert not toggle.is_on()
    assert toggle.is_off()


async def test_initialised_as_off():
    toggle = Toggle(False)

    assert not toggle
    assert not toggle.is_on()
    assert toggle.is_off()


async def test_initialised_as_on():
    toggle = Toggle(True)

    assert toggle
    assert toggle.is_on()
    assert not toggle.is_off()


async def test_turning_on():
    toggle = Toggle(False)
    await toggle.turn_on()

    assert toggle
    assert toggle.is_on()
    assert not toggle.is_off()


async def test_turning_off():
    toggle = Toggle(True)
    await toggle.turn_off()

    assert not toggle
    assert not toggle.is_on()
    assert toggle.is_off()


async def test_waiting_until_on_fails_when_not_turned_on():
    toggle = Toggle(False)

    with pytest.raises(asyncio.TimeoutError):
        await asyncio.wait_for(toggle.wait_for_on(), timeout=0.1)

    assert not toggle


async def test_waiting_until_off_fails_when_not_turned_off():
    toggle = Toggle(True)

    with pytest.raises(asyncio.TimeoutError):
        await asyncio.wait_for(toggle.wait_for_off(), timeout=0.1)

    assert toggle


async def test_waiting_until_on_wakes_when_turned_on(timer):
    toggle = Toggle(False)

    async def delayed_turning_on(delay: float):
        await asyncio.sleep(delay)
        await toggle.turn_on()

    with timer:
        asyncio.create_task(delayed_turning_on(0.05))
        await asyncio.wait_for(toggle.wait_for_on(), timeout=1.0)

    assert toggle
    assert timer.seconds < 0.5  # approx. 0.05 plus some code overhead


async def test_waiting_until_off_wakes_when_turned_off(timer):
    toggle = Toggle(True)

    async def delayed_turning_off(delay: float):
        await asyncio.sleep(delay)
        await toggle.turn_off()

    with timer:
        asyncio.create_task(delayed_turning_off(0.05))
        await asyncio.wait_for(toggle.wait_for_off(), timeout=1.0)

    assert not toggle
    assert timer.seconds < 0.5  # approx. 0.05 plus some code overhead



import datetime

import freezegun
import pytest

from kopf.engines.peering import CLUSTER_PEERING_RESOURCE, LEGACY_PEERING_RESOURCE, \
                                 NAMESPACED_PEERING_RESOURCE, Peer


@freezegun.freeze_time('2020-12-31T23:59:59.123456')
def test_defaults():
    peer = Peer(id='id', name='name')
    assert peer.id == 'id'
    assert peer.name == 'name'
    assert peer.namespace is None
    assert peer.legacy is False
    assert peer.lifetime == datetime.timedelta(seconds=60)
    assert peer.lastseen == datetime.datetime(2020, 12, 31, 23, 59, 59, 123456)


@freezegun.freeze_time('2020-12-31T23:59:59.123456')
def test_repr():
    peer = Peer(id='some-id', name='some-name', namespace='some-namespace')
    text = repr(peer)
    assert text.startswith('Peer(')
    assert text.endswith(')')
    assert '(some-id, ' in text
    assert 'priority=0' in text
    assert 'lastseen=' in text
    assert 'lifetime=' in text

    # The peering object's name is of no interest, the peer's id is.
    assert 'name=' not in text

    # The namespace of the operator can affect the conflict detection.
    # It is not always the same as the peering object's namespace.
    assert 'namespace=some-namespace' in text


@freezegun.freeze_time('2020-12-31T23:59:59.123456')
def test_priority_specified():
    peer = Peer(id='id', name='name', priority=123)
    assert peer.priority == 123


@freezegun.freeze_time('2020-12-31T23:59:59.123456')
def test_priority_unspecified():
    peer = Peer(id='id', name='name')
    assert peer.priority == 0


@pytest.mark.parametrize('namespace', [None, 'namespaced'])
def test_resource_for_legacy_peering(namespace):
    peer = Peer(id='id', name='name', legacy=True, namespace=namespace)
    assert peer.legacy is True
    assert peer.resource == LEGACY_PEERING_RESOURCE


def test_resource_for_cluster_peering():
    peer = Peer(id='id', name='name', legacy=False, namespace=None)
    assert peer.legacy is False
    assert peer.resource == CLUSTER_PEERING_RESOURCE
    assert peer.namespace is None


def test_resource_for_namespaced_peering():
    peer = Peer(id='id', name='name', legacy=False, namespace='namespaced')
    assert peer.legacy is False
    assert peer.resource == NAMESPACED_PEERING_RESOURCE
    assert peer.namespace == 'namespaced'


@freezegun.freeze_time('2020-12-31T23:59:59.123456')
def test_creation_with_lifetime_as_timedelta():
    peer = Peer(id='id', name='name', lifetime=datetime.timedelta(seconds=123))
    assert peer.lifetime == datetime.timedelta(seconds=123)


@freezegun.freeze_time('2020-12-31T23:59:59.123456')
def test_creation_with_lifetime_as_number():
    peer = Peer(id='id', name='name', lifetime=123)
    assert peer.lifetime == datetime.timedelta(seconds=123)


@freezegun.freeze_time('2020-12-31T23:59:59.123456')
def test_creation_with_lifetime_unspecified():
    peer = Peer(id='id', name='name')
    assert peer.lifetime == datetime.timedelta(seconds=60)


@freezegun.freeze_time('2020-12-31T23:59:59.123456')
def test_creation_with_lastseen_as_datetime():
    peer = Peer(id='id', name='name', lastseen=datetime.datetime(2020, 1, 1, 12, 34, 56, 789123))
    assert peer.lastseen == datetime.datetime(2020, 1, 1, 12, 34, 56, 789123)


@freezegun.freeze_time('2020-12-31T23:59:59.123456')
def test_creation_with_lastseen_as_string():
    peer = Peer(id='id', name='name', lastseen='2020-01-01T12:34:56.789123')
    assert peer.lastseen == datetime.datetime(2020, 1, 1, 12, 34, 56, 789123)


@freezegun.freeze_time('2020-12-31T23:59:59.123456')
def test_creation_with_lastseen_unspecified():
    peer = Peer(id='id', name='name')
    assert peer.lastseen == datetime.datetime(2020, 12, 31, 23, 59, 59, 123456)


@freezegun.freeze_time('2020-12-31T23:59:59.123456')
def test_creation_as_alive():
    peer = Peer(
        id='id',
        name='name',
        lifetime=10,
        lastseen='2020-12-31T23:59:50.123456',  # less than 10 seconds before "now"
    )
    assert peer.lifetime == datetime.timedelta(seconds=10)
    assert peer.lastseen == datetime.datetime(2020, 12, 31, 23, 59, 50, 123456)
    assert peer.deadline == datetime.datetime(2021, 1, 1, 0, 0, 0, 123456)
    assert peer.is_dead is False


@freezegun.freeze_time('2020-12-31T23:59:59.123456')
def test_creation_as_dead():
    peer = Peer(
        id='id',
        name='name',
        lifetime=10,
        lastseen='2020-12-31T23:59:49.123456',  # 10 seconds before "now" sharp
    )
    assert peer.lifetime == datetime.timedelta(seconds=10)
    assert peer.lastseen == datetime.datetime(2020, 12, 31, 23, 59, 49, 123456)
    assert peer.deadline == datetime.datetime(2020, 12, 31, 23, 59, 59, 123456)
    assert peer.is_dead is True


def test_touching_when_alive():
    with freezegun.freeze_time('2020-01-01T10:20:30'):
        peer = Peer(id='id1', name='name1', lifetime=123)

    assert not peer.is_dead

    with freezegun.freeze_time('2020-02-02T11:22:33'):
        peer.touch()

    assert peer.lifetime == datetime.timedelta(seconds=123)
    assert peer.lastseen == datetime.datetime(2020, 2, 2, 11, 22, 33)
    assert peer.deadline == datetime.datetime(2020, 2, 2, 11, 24, 36)
    assert not peer.is_dead


def test_touching_when_dead():
    with freezegun.freeze_time('2020-01-01T10:20:30'):
        peer = Peer(id='id1', name='name1', lifetime=123, lastseen='2019-01-01T00:00:00')

    assert peer.is_dead

    with freezegun.freeze_time('2020-02-02T11:22:33'):
        peer.touch()

    assert peer.lifetime == datetime.timedelta(seconds=123)
    assert peer.lastseen == datetime.datetime(2020, 2, 2, 11, 22, 33)
    assert peer.deadline == datetime.datetime(2020, 2, 2, 11, 24, 36)
    assert not peer.is_dead



"""
The main Kopf module for all the exported functions & classes.
"""
# isort: skip_file

# Unlike all other places, where we import other modules and refer
# the functions via the modules, this is the framework's top-level interface,
# as it is seen by the users. So, we export the individual functions.

from kopf import (
    on,  # as a separate name on the public namespace
)
from kopf.config import (
    LOGLEVEL_INFO,  # deprecated
    LOGLEVEL_WARNING,  # deprecated
    LOGLEVEL_ERROR,  # deprecated
    LOGLEVEL_CRITICAL,  # deprecated
    EventsConfig,  # deprecated
    WorkersConfig,  # deprecated
)
from kopf.engines.loggers import (
    configure,
    LogFormat,
    ObjectLogger,
    LocalObjectLogger,
)
from kopf.engines.posting import (
    event,
    info,
    warn,
    exception,
)
from kopf.on import (
    daemon,
    timer,
)
from kopf.on import (
    register,
)
from kopf.reactor import (
    lifecycles,  # as a separate name on the public namespace
)
from kopf.reactor.handling import (
    TemporaryError,
    PermanentError,
    HandlerTimeoutError,
    HandlerRetriesError,
    execute,
)
from kopf.reactor.lifecycles import (
    get_default_lifecycle,
    set_default_lifecycle,
)
from kopf.reactor.registries import (
    ActivityRegistry,
    ResourceRegistry,
    ResourceWatchingRegistry,
    ResourceChangingRegistry,
    OperatorRegistry,
    get_default_registry,
    set_default_registry,
)
from kopf.reactor.running import (
    spawn_tasks,
    run_tasks,
    operator,
    run,
    login,  # deprecated
    create_tasks,  # deprecated
)
from kopf.storage.diffbase import (
    DiffBaseStorage,
    AnnotationsDiffBaseStorage,
    StatusDiffBaseStorage,
    MultiDiffBaseStorage,
)
from kopf.storage.progress import (
    ProgressRecord,
    ProgressStorage,
    AnnotationsProgressStorage,
    StatusProgressStorage,
    MultiProgressStorage,
    SmartProgressStorage,
)
from kopf.structs.bodies import (
    RawEventType,
    RawEvent,
    RawBody,
    Status,
    Spec,
    Meta,
    Body,
    BodyEssence,
    OwnerReference,
    ObjectReference,
    build_object_reference,
    build_owner_reference,
)
from kopf.structs.callbacks import (
    not_,
    all_,
    any_,
    none_,
)
from kopf.structs.configuration import (
    OperatorSettings,
)
from kopf.structs.containers import (
    Memo,
)
from kopf.structs.credentials import (
    LoginError,
    ConnectionInfo,
)
from kopf.structs.dicts import (
    FieldSpec,
    FieldPath,
)
from kopf.structs.diffs import (
    Diff,
    DiffItem,
    DiffOperation,
)
from kopf.structs.filters import (
    ABSENT,
    PRESENT,
)
from kopf.structs.handlers import (
    HandlerId,
    ErrorsMode,
    Reason,
)
from kopf.structs.patches import (
    Patch,
)
from kopf.structs.primitives import (
    DaemonStoppingReason,
    SyncDaemonStopperChecker,
    AsyncDaemonStopperChecker,
)
from kopf.structs.resources import (
    Resource,
)
from kopf.toolkits.hierarchies import (
    adopt,
    label,
    harmonize_naming,
    adjust_namespace,
    append_owner_reference,
    remove_owner_reference,
)
from kopf.toolkits.legacy_registries import (
    BaseRegistry,
    SimpleRegistry,
    GlobalRegistry,
)
from kopf.utilities.piggybacking import (
    login_via_pykube,
    login_via_client,
)

HandlerFatalError = PermanentError  # a backward-compatibility alias
HandlerRetryError = TemporaryError  # a backward-compatibility alias

__all__ = [
    'on', 'lifecycles', 'register', 'execute', 'daemon', 'timer',
    'configure', 'LogFormat',
    'login', 'LoginError', 'ConnectionInfo',
    'login_via_pykube', 'login_via_client',
    'event', 'info', 'warn', 'exception',
    'spawn_tasks', 'run_tasks', 'operator', 'run', 'create_tasks',
    'adopt', 'label',
    'not_',
    'all_',
    'any_',
    'none_',
    'get_default_lifecycle', 'set_default_lifecycle',
    'build_object_reference', 'build_owner_reference',
    'append_owner_reference', 'remove_owner_reference',
    'ErrorsMode',
    'PermanentError', 'HandlerFatalError',
    'TemporaryError', 'HandlerRetryError',
    'HandlerTimeoutError',
    'HandlerRetriesError',
    'BaseRegistry',  # deprecated
    'SimpleRegistry',  # deprecated
    'GlobalRegistry',  # deprecated
    'ActivityRegistry',
    'ResourceRegistry',
    'ResourceWatchingRegistry',
    'ResourceChangingRegistry',
    'OperatorRegistry',
    'get_default_registry',
    'set_default_registry',
    'PRESENT', 'ABSENT',
    'OperatorSettings',
    'DiffBaseStorage',
    'AnnotationsDiffBaseStorage',
    'StatusDiffBaseStorage',
    'MultiDiffBaseStorage',
    'ProgressRecord',
    'ProgressStorage',
    'AnnotationsProgressStorage',
    'StatusProgressStorage',
    'MultiProgressStorage',
    'SmartProgressStorage',
    'DaemonStoppingReason',
    'RawEventType',
    'RawEvent',
    'RawBody',
    'Status',
    'Spec',
    'Meta',
    'Body',
    'BodyEssence',
    'ObjectReference',
    'OwnerReference',
    'Memo',
    'ObjectLogger',
    'LocalObjectLogger',
    'FieldSpec',
    'FieldPath',
    'Diff',
    'DiffItem',
    'DiffOperation',
    'HandlerId',
    'Reason',
    'Patch',
    'SyncDaemonStopperChecker',
    'AsyncDaemonStopperChecker',
    'Resource',
]



from typing import Optional, cast

import aiohttp

from kopf.clients import auth, discovery
from kopf.structs import bodies, patches, resources


@auth.reauthenticated_request
async def patch_obj(
        *,
        resource: resources.Resource,
        patch: patches.Patch,
        namespace: Optional[str] = None,
        name: Optional[str] = None,
        body: Optional[bodies.Body] = None,
        context: Optional[auth.APIContext] = None,  # injected by the decorator
) -> bodies.RawBody:
    """
    Patch a resource of specific kind.

    Either the namespace+name should be specified, or the body,
    which is used only to get namespace+name identifiers.

    Unlike the object listing, the namespaced call is always
    used for the namespaced resources, even if the operator serves
    the whole cluster (i.e. is not namespace-restricted).

    Returns the patched body. The patched body can be partial (status-only,
    no-status, or empty) -- depending on whether there were fields in the body
    or in the status to patch; if neither had fields for patching, the result
    is an empty body. The result should only be used to check against the patch:
    if there was nothing to patch, it does not matter if the fields are absent.
    """
    if context is None:
        raise RuntimeError("API instance is not injected by the decorator.")

    if body is not None and (name is not None or namespace is not None):
        raise TypeError("Either body, or name+namespace can be specified. Got both.")

    namespace = body.get('metadata', {}).get('namespace') if body is not None else namespace
    name = body.get('metadata', {}).get('name') if body is not None else name

    is_namespaced = await discovery.is_namespaced(resource=resource, context=context)
    namespace = namespace if is_namespaced else None

    if body is None:
        body = cast(bodies.Body, {'metadata': {'name': name}})
        if namespace is not None:
            body['metadata']['namespace'] = namespace

    as_subresource = await discovery.is_status_subresource(resource=resource, context=context)
    body_patch = dict(patch)  # shallow: for mutation of the top-level keys below.
    status_patch = body_patch.pop('status', None) if as_subresource else None

    # Patch & reconstruct the actual body as reported by the server. The reconstructed body can be
    # partial or empty -- if the body/status patches are empty. This is fine: it is only used
    # to verify that the patched fields are matching the patch. No patch? No mismatch!
    patched_body = bodies.RawBody()
    try:
        if body_patch:
            response = await context.session.patch(
                url=resource.get_url(server=context.server, namespace=namespace, name=name),
                headers={'Content-Type': 'application/merge-patch+json'},
                json=body_patch,
                raise_for_status=True,
            )
            patched_body = await response.json()

        if status_patch:
            response = await context.session.patch(
                url=resource.get_url(server=context.server, namespace=namespace, name=name,
                                     subresource='status' if as_subresource else None),
                headers={'Content-Type': 'application/merge-patch+json'},
                json={'status': status_patch},
                raise_for_status=True,
            )
            patched_body['status'] = await response.json()

    except aiohttp.ClientResponseError as e:
        if e.status == 404:
            pass
        else:
            raise

    return patched_body



import collections.abc

import pytest

from kopf import GlobalRegistry, SimpleRegistry


# Used in the tests. Must be global-scoped, or its qualname will be affected.
def some_fn():
    pass


def test_simple_registry_via_iter(cause_factory):

    cause = cause_factory()
    registry = SimpleRegistry()
    iterator = registry.iter_cause_handlers(cause)

    assert isinstance(iterator, collections.abc.Iterator)
    assert not isinstance(iterator, collections.abc.Collection)
    assert not isinstance(iterator, collections.abc.Container)
    assert not isinstance(iterator, (list, tuple))

    with pytest.deprecated_call(match=r"use ResourceChangingRegistry.iter_handlers\(\)"):
        handlers = list(iterator)
    assert not handlers


def test_simple_registry_via_list(cause_factory):

    cause = cause_factory()
    registry = SimpleRegistry()
    with pytest.deprecated_call(match=r"use ResourceChangingRegistry.get_handlers\(\)"):
        handlers = registry.get_cause_handlers(cause)

    assert isinstance(handlers, collections.abc.Iterable)
    assert isinstance(handlers, collections.abc.Container)
    assert isinstance(handlers, collections.abc.Collection)
    assert not handlers


def test_simple_registry_with_minimal_signature(cause_factory):

    cause = cause_factory()
    registry = SimpleRegistry()
    with pytest.deprecated_call(match=r"registry.register\(\) is deprecated"):
        registry.register(some_fn)
    with pytest.deprecated_call(match=r"use ResourceChangingRegistry.get_handlers\(\)"):
        handlers = registry.get_cause_handlers(cause)

    assert len(handlers) == 1
    assert handlers[0].fn is some_fn


def test_global_registry_via_iter(cause_factory):

    cause = cause_factory()
    registry = GlobalRegistry()
    iterator = registry.iter_cause_handlers(cause)

    assert isinstance(iterator, collections.abc.Iterator)
    assert not isinstance(iterator, collections.abc.Collection)
    assert not isinstance(iterator, collections.abc.Container)
    assert not isinstance(iterator, (list, tuple))

    with pytest.deprecated_call(match=r"use OperatorRegistry.iter_resource_changing_handlers\(\)"):
        handlers = list(iterator)
    assert not handlers


def test_global_registry_via_list(cause_factory):

    cause = cause_factory()
    registry = GlobalRegistry()
    with pytest.deprecated_call(match=r"use OperatorRegistry.get_resource_changing_handlers\(\)"):
        handlers = registry.get_cause_handlers(cause)

    assert isinstance(handlers, collections.abc.Iterable)
    assert isinstance(handlers, collections.abc.Container)
    assert isinstance(handlers, collections.abc.Collection)
    assert not handlers


def test_global_registry_with_minimal_signature(cause_factory, resource):

    cause = cause_factory()
    registry = GlobalRegistry()
    with pytest.deprecated_call(match=r"use OperatorRegistry.register_resource_changing_handler\(\)"):
        registry.register_cause_handler(resource.group, resource.version, resource.plural, some_fn)
    with pytest.deprecated_call(match=r"use OperatorRegistry.get_resource_changing_handlers\(\)"):
        handlers = registry.get_cause_handlers(cause)

    assert len(handlers) == 1
    assert handlers[0].fn is some_fn




import aiohttp.web
import freezegun
import pytest

from kopf.engines.peering import CLUSTER_PEERING_RESOURCE, \
                                 NAMESPACED_PEERING_RESOURCE, Peer, apply_peers


@pytest.mark.usefixtures('with_both_crds')
@pytest.mark.parametrize('namespace, peering_resource', [
    pytest.param('ns', NAMESPACED_PEERING_RESOURCE, id='namespace-scoped'),
    pytest.param(None, CLUSTER_PEERING_RESOURCE, id='cluster-scoped'),
])
@freezegun.freeze_time('2020-12-31T23:59:59.123456')
async def test_applying_a_dead_peer_purges_it(
        hostname, aresponses, resp_mocker, namespace, peering_resource):

    patch_mock = resp_mocker(return_value=aiohttp.web.json_response({}))
    url = peering_resource.get_url(name='name0', namespace=namespace)
    aresponses.add(hostname, url, 'patch', patch_mock)

    peer = Peer(id='id1', name='...', namespace='ns1', lastseen='2020-01-01T00:00:00')
    await apply_peers(peers=[peer], name='name0', namespace=namespace)

    assert patch_mock.called
    patch = await patch_mock.call_args_list[0][0][0].json()
    assert set(patch['status']) == {'id1'}
    assert patch['status']['id1'] is None


@pytest.mark.usefixtures('with_both_crds')
@pytest.mark.parametrize('namespace, peering_resource', [
    pytest.param('ns', NAMESPACED_PEERING_RESOURCE, id='namespace-scoped'),
    pytest.param(None, CLUSTER_PEERING_RESOURCE, id='cluster-scoped'),
])
@freezegun.freeze_time('2020-12-31T23:59:59.123456')
async def test_applying_an_alive_peer_stores_it(
        hostname, aresponses, resp_mocker, namespace, peering_resource):

    patch_mock = resp_mocker(return_value=aiohttp.web.json_response({}))
    url = peering_resource.get_url(name='name0', namespace=namespace)
    aresponses.add(hostname, url, 'patch', patch_mock)

    peer = Peer(id='id1', name='...', namespace='ns1', lastseen='2020-12-31T23:59:59')
    await apply_peers(peers=[peer], name='name0', namespace=namespace)

    assert patch_mock.called
    patch = await patch_mock.call_args_list[0][0][0].json()
    assert set(patch['status']) == {'id1'}
    assert patch['status']['id1']['namespace'] == 'ns1'
    assert patch['status']['id1']['priority'] == 0
    assert patch['status']['id1']['lastseen'] == '2020-12-31T23:59:59'
    assert patch['status']['id1']['lifetime'] == 60


@pytest.mark.usefixtures('with_both_crds')
@pytest.mark.parametrize('namespace, peering_resource', [
    pytest.param('ns', NAMESPACED_PEERING_RESOURCE, id='namespace-scoped'),
    pytest.param(None, CLUSTER_PEERING_RESOURCE, id='cluster-scoped'),
])
@pytest.mark.parametrize('lastseen', [
    pytest.param('2020-01-01T00:00:00', id='when-dead'),
    pytest.param('2020-12-31T23:59:59', id='when-alive'),
])
@freezegun.freeze_time('2020-12-31T23:59:59.123456')
async def test_keepalive(
        hostname, aresponses, resp_mocker, namespace, peering_resource, lastseen):

    patch_mock = resp_mocker(return_value=aiohttp.web.json_response({}))
    url = peering_resource.get_url(name='name0', namespace=namespace)
    aresponses.add(hostname, url, 'patch', patch_mock)

    peer = Peer(id='id1', name='name0', namespace=namespace, lastseen=lastseen)
    await peer.keepalive()

    assert patch_mock.called
    patch = await patch_mock.call_args_list[0][0][0].json()
    assert set(patch['status']) == {'id1'}
    assert patch['status']['id1']['namespace'] == namespace
    assert patch['status']['id1']['priority'] == 0
    assert patch['status']['id1']['lastseen'] == '2020-12-31T23:59:59.123456'
    assert patch['status']['id1']['lifetime'] == 60


@pytest.mark.usefixtures('with_both_crds')
@pytest.mark.parametrize('namespace, peering_resource', [
    pytest.param('ns', NAMESPACED_PEERING_RESOURCE, id='namespace-scoped'),
    pytest.param(None, CLUSTER_PEERING_RESOURCE, id='cluster-scoped'),
])
@pytest.mark.parametrize('lastseen', [
    pytest.param('2020-01-01T00:00:00', id='when-dead'),
    pytest.param('2020-12-31T23:59:59', id='when-alive'),
])
@freezegun.freeze_time('2020-12-31T23:59:59.123456')
async def test_disappear(
        hostname, aresponses, resp_mocker, namespace, peering_resource, lastseen):

    patch_mock = resp_mocker(return_value=aiohttp.web.json_response({}))
    url = peering_resource.get_url(name='name0', namespace=namespace)
    aresponses.add(hostname, url, 'patch', patch_mock)

    peer = Peer(id='id1', name='name0', namespace=namespace, lastseen=lastseen)
    await peer.disappear()

    assert patch_mock.called
    patch = await patch_mock.call_args_list[0][0][0].json()
    assert set(patch['status']) == {'id1'}
    assert patch['status']['id1'] is None



"""
Watching and streaming watch-events.

Kubernetes client's watching streams are synchronous. To make them asynchronous,
we put them into a `concurrent.futures.ThreadPoolExecutor`,
and yield from there asynchronously.

However, async/await coroutines misbehave with `StopIteration` exceptions
raised by the `next` method: see `PEP-479`_.

As a workaround, we replace `StopIteration` with our custom `StopStreaming`
inherited from `RuntimeError` (as suggested by `PEP-479`_),
and re-implement the generators to make them async.

All of this is a workaround for the standard Kubernetes client's limitations.
They would not be needed if the client library were natively asynchronous.

.. _PEP-479: https://www.python.org/dev/peps/pep-0479/
"""

import asyncio
import contextlib
import json
import logging
from typing import TYPE_CHECKING, Any, AsyncIterator, Dict, Optional, cast

import aiohttp

from kopf.clients import auth, discovery, fetching
from kopf.structs import bodies, configuration, primitives, resources
from kopf.utilities import backports

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    asyncio_Future = asyncio.Future[Any]
else:
    asyncio_Future = asyncio.Future


class WatchingError(Exception):
    """
    Raised when an unexpected error happens in the watch-stream API.
    """


async def infinite_watch(
        *,
        settings: configuration.OperatorSettings,
        resource: resources.Resource,
        namespace: Optional[str],
        freeze_mode: Optional[primitives.Toggle] = None,
) -> AsyncIterator[bodies.RawEvent]:
    """
    Stream the watch-events infinitely.

    This routine is extracted because it is difficult to test infinite loops.
    It is made as simple as possible, and is assumed to work without testing.

    This routine never ends gracefully. If a watcher's stream fails,
    a new one is recreated, and the stream continues.
    It only exits with unrecoverable exceptions.
    """
    while True:
        stream = streaming_watch(
            settings=settings,
            resource=resource,
            namespace=namespace,
            freeze_mode=freeze_mode,
        )
        async for raw_event in stream:
            yield raw_event
        await asyncio.sleep(settings.watching.reconnect_backoff)


async def streaming_watch(
        *,
        settings: configuration.OperatorSettings,
        resource: resources.Resource,
        namespace: Optional[str],
        freeze_mode: Optional[primitives.Toggle] = None,
) -> AsyncIterator[bodies.RawEvent]:

    # Prevent both watching and listing while the freeze mode is on, until it is off.
    # Specifically, the watch-stream closes its connection once the freeze mode is on,
    # so the while-true & for-event-in-stream cycles exit, and this coroutine is started
    # again by the `infinite_stream()` (the watcher timeout is swallowed by the freeze time).
    if freeze_mode is not None and freeze_mode.is_on():
        logger.debug("Freezing the watch-stream for %r", resource)
        await freeze_mode.wait_for_off()
        logger.debug("Resuming the watch-stream for %r", resource)

    # A stop-feature is a client-specific way of terminating the streaming HTTPS connection
    # when a freeze-mode is turned on. The low-level API call attaches its `response.close()`
    # to the future's callbacks, and a background task triggers it when the mode is turned on.
    freeze_waiter: asyncio_Future
    if freeze_mode is not None:
        freeze_waiter = backports.create_task(
            freeze_mode.wait_for_on(),
            name=f'freeze-waiter for {resource.name} @ {namespace or "cluster-wide"}')
    else:
        freeze_waiter = asyncio.Future()  # a dummy just ot have it

    try:
        stream = continuous_watch(
            settings=settings,
            resource=resource, namespace=namespace,
            freeze_waiter=freeze_waiter,
        )
        async for raw_event in stream:
            yield raw_event
    finally:
        with contextlib.suppress(asyncio.CancelledError):
            freeze_waiter.cancel()
            await freeze_waiter


async def continuous_watch(
        *,
        settings: configuration.OperatorSettings,
        resource: resources.Resource,
        namespace: Optional[str],
        freeze_waiter: asyncio_Future,
) -> AsyncIterator[bodies.RawEvent]:

    # First, list the resources regularly, and get the list's resource version.
    # Simulate the events with type "None" event - used in detection of causes.
    items, resource_version = await fetching.list_objs_rv(resource=resource, namespace=namespace)
    for item in items:
        yield {'type': None, 'object': item}

    # Repeat through disconnects of the watch as long as the resource version is valid (no errors).
    # The individual watching API calls are disconnected by timeout even if the stream is fine.
    while not freeze_waiter.done():

        # Then, watch the resources starting from the list's resource version.
        stream = watch_objs(
            settings=settings,
            resource=resource, namespace=namespace,
            timeout=settings.watching.server_timeout,
            since=resource_version,
            freeze_waiter=freeze_waiter,
        )
        async for raw_input in stream:
            raw_type = raw_input['type']
            raw_object = raw_input['object']

            # "410 Gone" is for the "resource version too old" error, we must restart watching.
            # The resource versions are lost by k8s after few minutes (5, as per the official doc).
            # The error occurs when there is nothing happening for few minutes. This is normal.
            if raw_type == 'ERROR' and cast(bodies.RawError, raw_object)['code'] == 410:
                logger.debug("Restarting the watch-stream for %r", resource)
                return  # out of the regular stream, to the infinite stream.

            # Other watch errors should be fatal for the operator.
            if raw_type == 'ERROR':
                raise WatchingError(f"Error in the watch-stream: {raw_object}")

            # Ensure that the event is something we understand and can handle.
            if raw_type not in ['ADDED', 'MODIFIED', 'DELETED']:
                logger.warning("Ignoring an unsupported event type: %r", raw_input)
                continue

            # Keep the latest seen resource version for continuation of the stream on disconnects.
            body = cast(bodies.RawBody, raw_object)
            resource_version = body.get('metadata', {}).get('resourceVersion', resource_version)

            # Yield normal events to the consumer. Errors are already filtered out.
            yield cast(bodies.RawEvent, raw_input)


@auth.reauthenticated_stream
async def watch_objs(
        *,
        settings: configuration.OperatorSettings,
        resource: resources.Resource,
        namespace: Optional[str] = None,
        timeout: Optional[float] = None,
        since: Optional[str] = None,
        context: Optional[auth.APIContext] = None,  # injected by the decorator
        freeze_waiter: asyncio_Future,
) -> AsyncIterator[bodies.RawInput]:
    """
    Watch objects of a specific resource type.

    The cluster-scoped call is used in two cases:

    * The resource itself is cluster-scoped, and namespacing makes not sense.
    * The operator serves all namespaces for the namespaced custom resource.

    Otherwise, the namespace-scoped call is used:

    * The resource is namespace-scoped AND operator is namespaced-restricted.
    """
    if context is None:
        raise RuntimeError("API instance is not injected by the decorator.")

    is_namespaced = await discovery.is_namespaced(resource=resource, context=context)
    namespace = namespace if is_namespaced else None

    params: Dict[str, str] = {}
    params['watch'] = 'true'
    if since is not None:
        params['resourceVersion'] = since
    if timeout is not None:
        params['timeoutSeconds'] = str(timeout)

    # Stream the parsed events from the response until it is closed server-side,
    # or until it is closed client-side by the freeze-waiting future's callbacks.
    try:
        response = await context.session.get(
            url=resource.get_url(server=context.server, namespace=namespace, params=params),
            timeout=aiohttp.ClientTimeout(
                total=settings.watching.client_timeout,
                sock_connect=settings.watching.connect_timeout,
            ),
        )
        response.raise_for_status()

        response_close_callback = lambda _: response.close()
        freeze_waiter.add_done_callback(response_close_callback)
        try:
            async with response:
                async for line in _iter_jsonlines(response.content):
                    raw_input = cast(bodies.RawInput, json.loads(line.decode("utf-8")))
                    yield raw_input
        finally:
            freeze_waiter.remove_done_callback(response_close_callback)

    except (aiohttp.ClientConnectionError, aiohttp.ClientPayloadError, asyncio.TimeoutError):
        pass


async def _iter_jsonlines(
        content: aiohttp.StreamReader,
        chunk_size: int = 1024 * 1024,
) -> AsyncIterator[bytes]:
    """
    Iterate line by line over the response's content.

    Usage::

        async for line in _iter_lines(response.content):
            pass

    This is an equivalent of::

        async for line in response.content:
            pass

    Except that the aiohttp's line iteration fails if the accumulated buffer
    length is above 2**17 bytes, i.e. 128 KB (`aiohttp.streams.DEFAULT_LIMIT`
    for the buffer's low-watermark, multiplied by 2 for the high-watermark).
    Kubernetes secrets and other fields can be much longer, up to MBs in length.

    The chunk size of 1MB is an empirical guess for keeping the memory footprint
    reasonably low on huge amount of small lines (limited to 1 MB in total),
    while ensuring the near-instant reads of the huge lines (can be a problem
    with a small chunk size due to too many iterations).

    .. seealso::
        https://github.com/zalando-incubator/kopf/issues/275
    """

    # Minimize the memory footprint by keeping at most 2 copies of a yielded line in memory
    # (in the buffer and as a yielded value), and at most 1 copy of other lines (in the buffer).
    buffer = b''
    async for data in content.iter_chunked(chunk_size):
        buffer += data
        del data

        start = 0
        index = buffer.find(b'\n', start)
        while index >= 0:
            line = buffer[start:index]
            if line:
                yield line
            del line
            start = index + 1
            index = buffer.find(b'\n', start)

        if start > 0:
            buffer = buffer[start:]

    if buffer:
        yield buffer



import aiohttp.web
import pytest

from kopf.clients.fetching import read_obj


@pytest.mark.resource_clustered  # see `resp_mocker`
async def test_when_present_clustered(
        resp_mocker, aresponses, hostname, resource):

    get_mock = resp_mocker(return_value=aiohttp.web.json_response({'a': 'b'}))
    aresponses.add(hostname, resource.get_url(namespace=None, name='name1'), 'get', get_mock)

    crd = await read_obj(resource=resource, namespace=None, name='name1')
    assert crd == {'a': 'b'}

    assert get_mock.called
    assert get_mock.call_count == 1


async def test_when_present_namespaced(
    resp_mocker, aresponses, hostname, resource):

    get_mock = resp_mocker(return_value=aiohttp.web.json_response({'a': 'b'}))
    aresponses.add(hostname, resource.get_url(namespace='ns1', name='name1'), 'get', get_mock)

    crd = await read_obj(resource=resource, namespace='ns1', name='name1')
    assert crd == {'a': 'b'}

    assert get_mock.called
    assert get_mock.call_count == 1


@pytest.mark.parametrize('namespace', [None, 'ns1'], ids=['without-namespace', 'with-namespace'])
@pytest.mark.parametrize('status', [403, 404])
async def test_when_absent_with_no_default(
        resp_mocker, aresponses, hostname, resource, namespace, status):

    get_mock = resp_mocker(return_value=aresponses.Response(status=status, reason="boo!"))
    aresponses.add(hostname, resource.get_url(namespace=None, name='name1'), 'get', get_mock)
    aresponses.add(hostname, resource.get_url(namespace='ns1', name='name1'), 'get', get_mock)

    with pytest.raises(aiohttp.ClientResponseError) as e:
        await read_obj(resource=resource, namespace=namespace, name='name1')
    assert e.value.status == status


@pytest.mark.parametrize('default', [None, object()], ids=['none', 'object'])
@pytest.mark.parametrize('namespace', [None, 'ns1'], ids=['without-namespace', 'with-namespace'])
@pytest.mark.parametrize('status', [403, 404])
async def test_when_absent_with_default(
        resp_mocker, aresponses, hostname, resource, namespace, default, status):

    get_mock = resp_mocker(return_value=aresponses.Response(status=status, reason="boo!"))
    aresponses.add(hostname, resource.get_url(namespace=None, name='name1'), 'get', get_mock)
    aresponses.add(hostname, resource.get_url(namespace='ns1', name='name1'), 'get', get_mock)

    crd = await read_obj(resource=resource, namespace=namespace, name='name1', default=default)
    assert crd is default


@pytest.mark.parametrize('namespace', [None, 'ns1'], ids=['without-namespace', 'with-namespace'])
@pytest.mark.parametrize('status', [400, 401, 500, 666])
async def test_raises_api_error_despite_default(
        resp_mocker, aresponses, hostname, resource, namespace, status):

    get_mock = resp_mocker(return_value=aresponses.Response(status=status, reason="boo!"))
    aresponses.add(hostname, resource.get_url(namespace=None, name='name1'), 'get', get_mock)
    aresponses.add(hostname, resource.get_url(namespace='ns1', name='name1'), 'get', get_mock)

    with pytest.raises(aiohttp.ClientResponseError) as e:
        await read_obj(resource=resource, namespace=namespace, name='name1', default=object())
    assert e.value.status == status



import asyncio
import sys
from typing import TYPE_CHECKING, Any, Awaitable, Generator, Optional, TypeVar, Union

_T = TypeVar('_T')

if TYPE_CHECKING:
    asyncio_Task = asyncio.Task[Any]
else:
    asyncio_Task = asyncio.Task

# Accept `name=` always, but simulate it for Python 3.7 to do nothing.
if sys.version_info >= (3, 8):
    create_task = asyncio.create_task
else:
    def create_task(
            coro: Union[Generator[Any, None, _T], Awaitable[_T]],
            *,
            name: Optional[str] = None,  # noqa
    ) -> asyncio_Task:
        return asyncio.create_task(coro)



"""
Synchronisation primitives and helper functions.
"""
import asyncio
import concurrent.futures
import enum
import threading
import time
from typing import TYPE_CHECKING, Any, Optional, Union

if TYPE_CHECKING:
    asyncio_Future = asyncio.Future[Any]
else:
    asyncio_Future = asyncio.Future

Flag = Union[asyncio_Future, asyncio.Event, concurrent.futures.Future, threading.Event]


async def wait_flag(
        flag: Optional[Flag],
) -> Any:
    """
    Wait for a flag to be raised.

    Non-asyncio primitives are generally not our worry,
    but we support them for convenience.
    """
    if flag is None:
        pass
    elif isinstance(flag, asyncio.Future):
        return await flag
    elif isinstance(flag, asyncio.Event):
        return await flag.wait()
    elif isinstance(flag, concurrent.futures.Future):
        loop = asyncio.get_running_loop()
        return await loop.run_in_executor(None, flag.result)
    elif isinstance(flag, threading.Event):
        loop = asyncio.get_running_loop()
        return await loop.run_in_executor(None, flag.wait)
    else:
        raise TypeError(f"Unsupported type of a flag: {flag!r}")


async def raise_flag(
        flag: Optional[Flag],
) -> None:
    """
    Raise a flag.

    Non-asyncio primitives are generally not our worry,
    but we support them for convenience.
    """
    if flag is None:
        pass
    elif isinstance(flag, asyncio.Future):
        flag.set_result(None)
    elif isinstance(flag, asyncio.Event):
        flag.set()
    elif isinstance(flag, concurrent.futures.Future):
        flag.set_result(None)
    elif isinstance(flag, threading.Event):
        flag.set()
    else:
        raise TypeError(f"Unsupported type of a flag: {flag!r}")


def check_flag(
        flag: Optional[Flag],
) -> Optional[bool]:
    """
    Check if a flag is raised.
    """
    if flag is None:
        return None
    elif isinstance(flag, asyncio.Future):
        return flag.done()
    elif isinstance(flag, asyncio.Event):
        return flag.is_set()
    elif isinstance(flag, concurrent.futures.Future):
        return flag.done()
    elif isinstance(flag, threading.Event):
        return flag.is_set()
    else:
        raise TypeError(f"Unsupported type of a flag: {flag!r}")


class Toggle:
    """
    An synchronisation primitive that can be awaited both until set or cleared.

    For one-directional toggles, `asyncio.Event` is sufficient.
    But these events cannot be awaited until cleared.

    The bi-directional toggles are needed in some places in the code, such as
    in the population/depletion of a `Vault`, or as an operator's freeze-mode.
    """

    def __init__(
            self,
            __val: bool = False,
    ) -> None:
        super().__init__()
        self._condition = asyncio.Condition()
        self._state: bool = bool(__val)

    def __bool__(self) -> bool:
        """
        In the boolean context, a toggle evaluates to its current on/off state.

        An equivalent of `.is_on` / `.is_off`.
        """
        return bool(self._state)

    def is_on(self) -> bool:
        """ Check if the toggle is currently on (opposite of `.is_off`). """
        return bool(self._state)

    def is_off(self) -> bool:
        """ Check if the toggle is currently off (opposite of `.is_on`). """
        return bool(not self._state)

    async def turn_on(self) -> None:
        """ Turn the toggle on, and wake up the tasks waiting for that. """
        async with self._condition:
            self._state = True
            self._condition.notify_all()

    async def turn_off(self) -> None:
        """ Turn the toggle off, and wake up the tasks waiting for that. """
        async with self._condition:
            self._state = False
            self._condition.notify_all()

    async def wait_for_on(self) -> None:
        """ Wait until the toggle is turned on (if not yet). """
        async with self._condition:
            await self._condition.wait_for(lambda: self._state)

    async def wait_for_off(self) -> None:
        """ Wait until the toggle is turned off (if not yet). """
        async with self._condition:
            await self._condition.wait_for(lambda: not self._state)


class DaemonStoppingReason(enum.Flag):
    """
    A reason or reasons of daemon being terminated.

    Daemons are signalled to exit usually for two reasons: the operator itself
    is exiting or restarting, so all daemons of all resources must stop;
    or the individual resource was deleted, but the operator continues running.

    No matter the reason, the daemons must exit, so one and only one stop-flag
    is used. Some daemons can check the reason of exiting if it is important.

    There can be multiple reasons combined (in rare cases, all of them).
    """
    NONE = 0
    DONE = enum.auto()  # whatever the reason and the status, the asyncio task has exited.
    FILTERS_MISMATCH = enum.auto()  # the resource does not match the filters anymore.
    RESOURCE_DELETED = enum.auto()  # the resource was deleted, the asyncio task is still awaited.
    OPERATOR_EXITING = enum.auto()  # the operator is exiting, the asyncio task is still awaited.
    DAEMON_SIGNALLED = enum.auto()  # the stopper flag was set, the asyncio task is still awaited.
    DAEMON_CANCELLED = enum.auto()  # the asyncio task was cancelled, the thread can be running.
    DAEMON_ABANDONED = enum.auto()  # we gave up on the asyncio task, the thread can be running.


class DaemonStopper:
    """
    A boolean flag indicating that the daemon should stop and exit.

    Every daemon gets a ``stopper`` kwarg, which is an event-like object.
    The stopper is raised in two cases:

    * The corresponding k8s object is deleted, so the daemon should stop.
    * The whole operator is stopping, so all the daemons should stop too.

    The stopper flag is a graceful way of a daemon termination.
    If the daemons do not react to their stoppers, and continue running,
    their tasks are cancelled by raising a `asyncio.CancelledError`.

    .. warning::
        In case of synchronous handlers, which are executed in the threads,
        this can lead to the OS resource leakage:
        there is no way to kill a thread in Python, so it will continue running
        forever or until failed (e.g. on an API call for an absent resource).
        The orphan threads will block the operator's process from exiting,
        thus affecting the speed of restarts.
    """

    def __init__(self) -> None:
        super().__init__()
        self.when: Optional[float] = None
        self.reason = DaemonStoppingReason.NONE
        self.sync_checker = SyncDaemonStopperChecker(self)
        self.async_checker = AsyncDaemonStopperChecker(self)
        self.sync_event = threading.Event()
        self.async_event = asyncio.Event()

    def __repr__(self) -> str:
        return f'<{self.__class__.__name__}: {self.is_set()}, reason={self.reason}>'

    def is_set(self, reason: Optional[DaemonStoppingReason] = None) -> bool:
        """
        Check if the daemon stopper is set: at all or for specific reason.
        """
        return ((reason is None or reason in self.reason) and self.sync_event.is_set())

    def set(self, *, reason: DaemonStoppingReason) -> None:
        self.when = self.when if self.when is not None else time.monotonic()
        self.reason |= reason
        self.sync_event.set()
        self.async_event.set()  # it is thread-safe: always called in operator's event loop.


class DaemonStopperChecker:

    """
    A minimalistic read-only checker for the daemons from the user side.

    This object is fed into the :kwarg:`stopped` kwarg for the handlers.

    The actual stopper is hidden from the users, and is an internal class.
    The users should not be able to trigger the stopping activities or
    check the reasons of stopping (or know about them at all).

    Usage::

        @kopf.daemon('zalando.org', 'v1', 'kopfexamples')
        def handler(stopped, **kwargs):
            while not stopped:
                ...
                stopped.wait(60)
    """

    def __init__(self, stopper: DaemonStopper) -> None:
        super().__init__()
        self._stopper = stopper

    def __repr__(self) -> str:
        return repr(self._stopper)

    def __bool__(self) -> bool:
        return self._stopper.is_set()

    def is_set(self) -> bool:
        return self._stopper.is_set()

    @property
    def reason(self) -> DaemonStoppingReason:
        return self._stopper.reason


class SyncDaemonStopperChecker(DaemonStopperChecker):
    def wait(self, timeout: Optional[float] = None) -> bool:
        self._stopper.sync_event.wait(timeout=timeout)
        return bool(self)


class AsyncDaemonStopperChecker(DaemonStopperChecker):
    async def wait(self, timeout: Optional[float] = None) -> bool:
        try:
            await asyncio.wait_for(self._stopper.async_event.wait(), timeout=timeout)
        except asyncio.TimeoutError:
            pass
        return bool(self)



from typing import Dict, Optional, cast

import aiohttp

from kopf.clients import auth
from kopf.structs import resources


@auth.reauthenticated_request
async def discover(
        *,
        resource: resources.Resource,
        subresource: Optional[str] = None,
        context: Optional[auth.APIContext] = None,  # injected by the decorator
) -> Optional[Dict[str, object]]:
    if context is None:
        raise RuntimeError("API instance is not injected by the decorator.")

    if resource.api_version not in context._discovered_resources:
        async with context._discovery_lock:
            if resource.api_version not in context._discovered_resources:
                context._discovered_resources[resource.api_version] = {}

                try:
                    response = await context.session.get(
                        url=resource.get_version_url(server=context.server),
                    )
                    response.raise_for_status()
                    respdata = await response.json()

                    context._discovered_resources[resource.api_version].update({
                        info['name']: info
                        for info in respdata['resources']
                    })

                except aiohttp.ClientResponseError as e:
                    if e.status in [403, 404]:
                        pass
                    else:
                        raise

    name = resource.plural if subresource is None else f'{resource.plural}/{subresource}'
    return context._discovered_resources[resource.api_version].get(name, None)


@auth.reauthenticated_request
async def is_namespaced(
        *,
        resource: resources.Resource,
        context: Optional[auth.APIContext] = None,  # injected by the decorator
) -> bool:
    if context is None:
        raise RuntimeError("API instance is not injected by the decorator.")

    info = await discover(resource=resource, context=context)
    return cast(bool, info['namespaced']) if info is not None else True  # assume namespaced


@auth.reauthenticated_request
async def is_status_subresource(
        *,
        resource: resources.Resource,
        context: Optional[auth.APIContext] = None,  # injected by the decorator
) -> bool:
    if context is None:
        raise RuntimeError("API instance is not injected by the decorator.")

    info = await discover(resource=resource, subresource='status', context=context)
    return info is not None



import collections
from unittest.mock import Mock

from kopf.reactor.registries import OperatorRegistry
from kopf.structs.resources import Resource


def test_resources():
    handler = Mock()

    resource1 = Resource('group1', 'version1', 'plural1')
    resource2 = Resource('group2', 'version2', 'plural2')

    registry = OperatorRegistry()
    registry.resource_watching_handlers[resource1].append(handler)
    registry.resource_changing_handlers[resource2].append(handler)
    registry.resource_watching_handlers[resource2].append(handler)
    registry.resource_changing_handlers[resource1].append(handler)

    resources = registry.resources

    assert isinstance(resources, collections.abc.Collection)
    assert len(resources) == 2

    assert resource1 in resources
    assert resource2 in resources



import enum
from typing import Mapping, Union

from kopf.structs import callbacks


class MetaFilterToken(enum.Enum):
    """ Tokens for filtering by annotations/labels. """
    PRESENT = enum.auto()
    ABSENT = enum.auto()


# For exporting to the top-level package.
ABSENT = MetaFilterToken.ABSENT
PRESENT = MetaFilterToken.PRESENT

# Filters for handler specifications (not the same as the object's values).
MetaFilter = Mapping[str, Union[None, str, MetaFilterToken, callbacks.MetaFilterFn]]



import kopf


def say_yes(value, spec, **_):
    return value == 'somevalue' and spec.get('field') is not None


def say_no(value, spec, **_):
    return value == 'somevalue' and spec.get('field') == 'not-this-value-for-sure'


@kopf.on.create('zalando.org', 'v1', 'kopfexamples', labels={'somelabel': 'somevalue'})
def create_with_labels_matching(logger, **kwargs):
    logger.info("Label is matching.")


@kopf.on.create('zalando.org', 'v1', 'kopfexamples', labels={'somelabel': kopf.PRESENT})
def create_with_labels_present(logger, **kwargs):
    logger.info("Label is present.")


@kopf.on.create('zalando.org', 'v1', 'kopfexamples', labels={'nonexistent': kopf.ABSENT})
def create_with_labels_absent(logger, **kwargs):
    logger.info("Label is absent.")


@kopf.on.create('zalando.org', 'v1', 'kopfexamples', labels={'somelabel': say_yes})
def create_with_labels_callback_matching(logger, **kwargs):
    logger.info("Label callback matching.")


@kopf.on.create('zalando.org', 'v1', 'kopfexamples', annotations={'someannotation': 'somevalue'})
def create_with_annotations_matching(logger, **kwargs):
    logger.info("Annotation is matching.")


@kopf.on.create('zalando.org', 'v1', 'kopfexamples', annotations={'someannotation': kopf.PRESENT})
def create_with_annotations_present(logger, **kwargs):
    logger.info("Annotation is present.")


@kopf.on.create('zalando.org', 'v1', 'kopfexamples', annotations={'nonexistent': kopf.ABSENT})
def create_with_annotations_absent(logger, **kwargs):
    logger.info("Annotation is absent.")


@kopf.on.create('zalando.org', 'v1', 'kopfexamples', annotations={'someannotation': say_no})
def create_with_annotations_callback_matching(logger, **kwargs):
    logger.info("Annotation callback mismatch.")


@kopf.on.create('zalando.org', 'v1', 'kopfexamples', when=lambda body, **_: True)
def create_with_filter_satisfied(logger, **kwargs):
    logger.info("Filter satisfied.")


@kopf.on.create('zalando.org', 'v1', 'kopfexamples', when=lambda body, **_: False)
def create_with_filter_not_satisfied(logger, **kwargs):
    logger.info("Filter not satisfied.")



import re

import pytest

from kopf.engines.peering import CLUSTER_PEERING_RESOURCE, NAMESPACED_PEERING_RESOURCE, \
                                 PEERING_DEFAULT_NAME, Peer

# Note: the legacy peering is intentionally not tested: it was long time before
# these tests were written, so it does not make sense to keep it stable.
# The legacy peering is going to be removed in version 1.0 when it happens.


@pytest.fixture()
def with_cluster_default(hostname, aresponses):
    url = CLUSTER_PEERING_RESOURCE.get_url(name=PEERING_DEFAULT_NAME)
    aresponses.add(hostname, url, 'get', {'spec': {}})


@pytest.fixture()
def with_cluster_specific(hostname, aresponses):
    url = CLUSTER_PEERING_RESOURCE.get_url(name='peering-name')
    aresponses.add(hostname, url, 'get', {'spec': {}})


@pytest.fixture()
def with_namespaced_default(hostname, aresponses):
    url = NAMESPACED_PEERING_RESOURCE.get_url(namespace='namespace', name=PEERING_DEFAULT_NAME)
    aresponses.add(hostname, url, 'get', {'spec': {}})


@pytest.fixture()
def with_namespaced_specific(hostname, aresponses):
    url = NAMESPACED_PEERING_RESOURCE.get_url(namespace='namespace', name='peering-name')
    aresponses.add(hostname, url, 'get', {'spec': {}})


#
# Parametrization via fixtures (it does not work from tests).
#
@pytest.fixture(params=[
    pytest.param(None, id='no-crds'),
    pytest.param('with_both_crds', id='both-crds'),
    pytest.param('with_cluster_crd', id='only-cluster-crd'),
    pytest.param('with_namespaced_crd', id='only-namespaced-crd'),
])
def all_crd_modes(request):
    return request.getfixturevalue(request.param) if request.param else None


@pytest.fixture(params=[
    pytest.param('with_both_crds', id='both-crds'),
    pytest.param('with_cluster_crd', id='only-cluster-crd'),
])
def all_crd_modes_with_cluster_scoped_crd(request):
    return request.getfixturevalue(request.param) if request.param else None


@pytest.fixture(params=[
    pytest.param('with_both_crds', id='both-crds'),
    pytest.param('with_namespaced_crd', id='only-namespaced-crd'),
])
def all_crd_modes_with_namespace_scoped_crd(request):
    return request.getfixturevalue(request.param) if request.param else None


@pytest.fixture(params=[
    pytest.param(True, id='with-cluster-default'),
    pytest.param(False, id='without-cluster-default')
])
def both_cluster_default_modes(request):
    return request.getfixturevalue('with_cluster_default') if request.param else None


@pytest.fixture(params=[
    pytest.param(True, id='with-cluster-specific'),
    pytest.param(False, id='without-cluster-specific')
])
def both_cluster_specific_modes(request):
    return request.getfixturevalue('with_cluster_specific') if request.param else None


@pytest.fixture(params=[
    pytest.param(True, id='with-namespaced-default'),
    pytest.param(False, id='without-namespaced-default')
])
def both_namespaced_default_modes(request):
    return request.getfixturevalue('with_namespaced_default') if request.param else None


@pytest.fixture(params=[
    pytest.param(True, id='with-namespaced-specific'),
    pytest.param(False, id='without-namespaced-specific')
])
def both_namespaced_specific_modes(request):
    return request.getfixturevalue('with_namespaced_specific') if request.param else None


#
# Actual tests: only the action & assertions.
#
@pytest.mark.usefixtures('both_namespaced_specific_modes')
@pytest.mark.usefixtures('both_namespaced_default_modes')
@pytest.mark.usefixtures('both_cluster_specific_modes')
@pytest.mark.usefixtures('both_cluster_default_modes')
@pytest.mark.usefixtures('all_crd_modes')
@pytest.mark.parametrize('name', [None, 'name'])
@pytest.mark.parametrize('namespace', [None, 'namespaced'])
async def test_standalone(namespace, name):
    peer = await Peer.detect(standalone=True, namespace=namespace, name=name)
    assert peer is None


@pytest.mark.usefixtures('both_namespaced_specific_modes')
@pytest.mark.usefixtures('both_namespaced_default_modes')
@pytest.mark.usefixtures('both_cluster_specific_modes')
@pytest.mark.usefixtures('with_cluster_default')
@pytest.mark.usefixtures('all_crd_modes_with_cluster_scoped_crd')
async def test_cluster_scoped_with_default_name():
    peer = await Peer.detect(id='id', standalone=False, namespace=None, name=None)
    assert peer.name == PEERING_DEFAULT_NAME
    assert peer.namespace is None


@pytest.mark.usefixtures('both_namespaced_specific_modes')
@pytest.mark.usefixtures('with_namespaced_default')
@pytest.mark.usefixtures('both_cluster_specific_modes')
@pytest.mark.usefixtures('both_cluster_default_modes')
@pytest.mark.usefixtures('all_crd_modes_with_namespace_scoped_crd')
async def test_namespace_scoped_with_default_name():
    peer = await Peer.detect(id='id', standalone=False, namespace='namespace', name=None)
    assert peer.name == PEERING_DEFAULT_NAME
    assert peer.namespace == 'namespace'


@pytest.mark.usefixtures('both_namespaced_specific_modes')
@pytest.mark.usefixtures('both_namespaced_default_modes')
@pytest.mark.usefixtures('with_cluster_specific')
@pytest.mark.usefixtures('both_cluster_default_modes')
@pytest.mark.usefixtures('all_crd_modes_with_cluster_scoped_crd')
async def test_cluster_scoped_with_specific_name():
    peer = await Peer.detect(id='id', standalone=False, namespace=None, name='peering-name')
    assert peer.name == 'peering-name'
    assert peer.namespace is None


@pytest.mark.usefixtures('with_namespaced_specific')
@pytest.mark.usefixtures('both_namespaced_default_modes')
@pytest.mark.usefixtures('both_cluster_specific_modes')
@pytest.mark.usefixtures('both_cluster_default_modes')
@pytest.mark.usefixtures('all_crd_modes_with_namespace_scoped_crd')
async def test_namespace_scoped_with_specific_name():
    peer = await Peer.detect(id='id', standalone=False, namespace='namespace', name='peering-name')
    assert peer.name == 'peering-name'
    assert peer.namespace == 'namespace'


@pytest.mark.usefixtures('both_namespaced_specific_modes')
@pytest.mark.usefixtures('both_namespaced_default_modes')
@pytest.mark.usefixtures('both_cluster_specific_modes')
@pytest.mark.usefixtures('both_cluster_default_modes')
@pytest.mark.usefixtures('all_crd_modes_with_cluster_scoped_crd')
async def test_cluster_scoped_with_absent_name(hostname, aresponses):
    aresponses.add(hostname, re.compile(r'.*'), 'get', aresponses.Response(status=404), repeat=999)
    with pytest.raises(Exception, match=r"The peering 'absent-name' was not found") as e:
        await Peer.detect(id='id', standalone=False, namespace=None, name='absent-name')


@pytest.mark.usefixtures('both_namespaced_specific_modes')
@pytest.mark.usefixtures('both_namespaced_default_modes')
@pytest.mark.usefixtures('both_cluster_specific_modes')
@pytest.mark.usefixtures('both_cluster_default_modes')
@pytest.mark.usefixtures('all_crd_modes_with_namespace_scoped_crd')
async def test_namespace_scoped_with_absent_name(hostname, aresponses):
    aresponses.add(hostname, re.compile(r'.*'), 'get', aresponses.Response(status=404), repeat=999)
    with pytest.raises(Exception, match=r"The peering 'absent-name' was not found") as e:
        await Peer.detect(id='id', standalone=False, namespace='namespace', name='absent-name')


# NB: without cluster-default peering.
@pytest.mark.usefixtures('both_namespaced_specific_modes')
@pytest.mark.usefixtures('both_namespaced_default_modes')
@pytest.mark.usefixtures('both_cluster_specific_modes')
@pytest.mark.usefixtures('all_crd_modes_with_cluster_scoped_crd')
async def test_cluster_scoped_with_warning(hostname, aresponses, assert_logs, caplog):
    aresponses.add(hostname, re.compile(r'.*'), 'get', aresponses.Response(status=404), repeat=999)
    peer = await Peer.detect(id='id', standalone=False, namespace=None, name=None)
    assert peer is None
    assert_logs([
        "Default peering object not found, falling back to the standalone mode."
    ])


# NB: without namespaced-default peering.
@pytest.mark.usefixtures('both_namespaced_specific_modes')
@pytest.mark.usefixtures('both_cluster_specific_modes')
@pytest.mark.usefixtures('both_cluster_default_modes')
@pytest.mark.usefixtures('all_crd_modes_with_namespace_scoped_crd')
async def test_namespace_scoped_with_warning(hostname, aresponses, assert_logs, caplog):
    aresponses.add(hostname, re.compile(r'.*'), 'get', aresponses.Response(status=404), repeat=999)
    peer = await Peer.detect(id='id', standalone=False, namespace='namespace', name=None)
    assert peer is None
    assert_logs([
        "Default peering object not found, falling back to the standalone mode."
    ])



import aiohttp.web
import pytest

from kopf.clients.discovery import discover, is_namespaced, is_status_subresource
from kopf.structs.resources import Resource


async def test_discovery_of_existing_resource(
        resp_mocker, aresponses, hostname):

    res1info = {'name': 'someresources', 'namespaced': True}
    result = {'resources': [res1info]}
    list_mock = resp_mocker(return_value=aiohttp.web.json_response(result))
    aresponses.add(hostname, '/apis/some-group.org/someversion', 'get', list_mock)

    resource = Resource('some-group.org', 'someversion', 'someresources')
    info = await discover(resource=resource)

    assert info == res1info


async def test_discovery_of_unexisting_resource(
        resp_mocker, aresponses, hostname):

    result = {'resources': []}
    list_mock = resp_mocker(return_value=aiohttp.web.json_response(result))
    aresponses.add(hostname, '/apis/some-group.org/someversion', 'get', list_mock)

    resource = Resource('some-group.org', 'someversion', 'someresources')
    info = await discover(resource=resource)

    assert info is None


@pytest.mark.parametrize('status', [403, 404])
async def test_discovery_of_unexisting_group_or_version(
        resp_mocker, aresponses, hostname, status):

    list_mock = resp_mocker(return_value=aresponses.Response(status=status, reason="boo!"))
    aresponses.add(hostname, '/apis/some-group.org/someversion', 'get', list_mock)

    resource = Resource('some-group.org', 'someversion', 'someresources')
    info = await discover(resource=resource)

    assert info is None


async def test_discovery_is_cached_per_session(
        resp_mocker, aresponses, hostname):

    res1info = {'name': 'someresources1', 'namespaced': True}
    res2info = {'name': 'someresources2', 'namespaced': True}

    result = {'resources': [res1info]}
    list_mock = resp_mocker(return_value=aiohttp.web.json_response(result))
    aresponses.add(hostname, '/apis/some-group.org/someversion', 'get', list_mock)

    result = {'resources': [res2info]}
    list_mock = resp_mocker(return_value=aiohttp.web.json_response(result))
    aresponses.add(hostname, '/apis/some-group.org/someversion', 'get', list_mock)

    resource = Resource('some-group.org', 'someversion', 'someresources1')
    info = await discover(resource=resource)
    assert info == res1info

    resource = Resource('some-group.org', 'someversion', 'someresources2')
    info = await discover(resource=resource)
    assert info is None  # cached as absent on the 1st call.

    resource = Resource('some-group.org', 'someversion', 'someresources1')
    info = await discover(resource=resource)
    assert info == res1info


@pytest.mark.parametrize('namespaced', [True, False])
async def test_is_namespaced(
        resp_mocker, aresponses, hostname, namespaced):

    res1info = {'name': 'someresources', 'namespaced': namespaced}
    result = {'resources': [res1info]}
    list_mock = resp_mocker(return_value=aiohttp.web.json_response(result))
    aresponses.add(hostname, '/apis/some-group.org/someversion', 'get', list_mock)

    resource = Resource('some-group.org', 'someversion', 'someresources')
    result = await is_namespaced(resource=resource)

    assert result == namespaced


@pytest.mark.parametrize('namespaced', [True, False])
async def test_is_status_subresource_when_not_a_subresource(
        resp_mocker, aresponses, hostname, namespaced):

    res1info = {'name': 'someresources', 'namespaced': namespaced}
    result = {'resources': [res1info]}
    list_mock = resp_mocker(return_value=aiohttp.web.json_response(result))
    aresponses.add(hostname, '/apis/some-group.org/someversion', 'get', list_mock)

    resource = Resource('some-group.org', 'someversion', 'someresources')
    result = await is_status_subresource(resource=resource)

    assert result is False  # an extra type-check


@pytest.mark.parametrize('namespaced', [True, False])
async def test_is_status_subresource_when_is_a_subresource(
        resp_mocker, aresponses, hostname, namespaced):

    res1info = {'name': 'someresources', 'namespaced': namespaced}
    res1status = {'name': 'someresources/status', 'namespaced': namespaced}
    result = {'resources': [res1info, res1status]}
    list_mock = resp_mocker(return_value=aiohttp.web.json_response(result))
    aresponses.add(hostname, '/apis/some-group.org/someversion', 'get', list_mock)

    resource = Resource('some-group.org', 'someversion', 'someresources')
    result = await is_status_subresource(resource=resource)

    assert result is True  # an extra type-check



import logging
from typing import Optional

# Deprecated: use ``logging.*`` constants instead. Kept here for backward-compatibility.
LOGLEVEL_INFO = logging.INFO
LOGLEVEL_WARNING = logging.WARNING
LOGLEVEL_ERROR = logging.ERROR
LOGLEVEL_CRITICAL = logging.CRITICAL


# DEPRECATED: Used for initial defaults for per-operator settings (see kopf.structs.configuration).
class EventsConfig:
    """
    Used to configure events sending behaviour.
    """

    events_loglevel: int = logging.INFO
    """ What events should be logged. """


# DEPRECATED: Used for initial defaults for per-operator settings (see kopf.structs.configuration).
class WorkersConfig:
    """
    Used as single point of configuration for kopf.reactor.
    """

    queue_workers_limit: Optional[int] = None  # if None, there is no limits to workers number
    """ How many workers can be running simultaneously on per-object event queue. """

    synchronous_tasks_threadpool_limit: Optional[int] = None  # if None, calculated by ThreadPoolExecutor based on cpu count
    """ How many threads in total can be running simultaneously to handle any non-async tasks."""

    worker_idle_timeout: float = 5.0
    """ How long does a worker can idle before exiting and garbage-collecting."""

    worker_batch_window: float = 0.1
    """ How fast/slow does a worker deplete the queue when an event is received."""

    worker_exit_timeout: float = 2.0
    """ How long does a worker can work on watcher exit before being cancelled. """

    @staticmethod
    def set_synchronous_tasks_threadpool_limit(new_limit: int) -> None:
        """
        Call this static method at any time to change synchronous_tasks_threadpool_limit in runtime.
        """
        if new_limit < 1:
            raise ValueError('Can`t set threadpool limit lower than 1')

        WorkersConfig.synchronous_tasks_threadpool_limit = new_limit

        # Also apply to the current runtime settings, if we are at runtime (not load-time).
        try:
            # Wherever we can find it; ignore "nice" architecture (this class is deprecated anyway).
            from kopf.engines import posting  # noqa
            from kopf.structs import configuration  # noqa  # cyclic imports, for type annotations
            settings: configuration.OperatorSettings = posting.settings_var.get()
        except LookupError:
            pass
        else:
            settings.execution.max_workers = new_limit


# DEPRECATED: Used for initial defaults for per-operator settings (see kopf.structs.configuration).
class WatchersConfig:
    """
    Used to configure the K8s API watchers and streams.
    """

    default_stream_timeout: Optional[float] = None
    """ The maximum duration of one streaming request. Patched in some tests. """

    watcher_retry_delay: float = 0.1
    """ How long should a pause be between watch requests (to prevent flooding). """

    session_timeout: Optional[float] = None
    """ The http session timeout to use in watch request. """



import os

import freezegun
import pytest

from kopf.engines.peering import detect_own_id


@pytest.fixture(autouse=True)
def _intercept_os_calls(mocker):
    mocker.patch('getpass.getuser', return_value='some-user')
    mocker.patch('socket.getfqdn', return_value='some-host')
    mocker.patch('random.choices', return_value='random-str')


def test_from_a_pod_id(mocker):
    mocker.patch.dict(os.environ, POD_ID='some-pod-1')
    own_id = detect_own_id()
    assert own_id == 'some-pod-1'


@freezegun.freeze_time('2020-12-31T23:59:59.123456')
def test_with_defaults():
    own_id = detect_own_id()
    assert own_id == 'some-user@some-host/2020-12-31T23:59:59.123456/random-str'



"""
Remember: We do not test the clients, we assume they work when used properly.
We test our own functions here, and check if the clients were called.
"""
import pykube
import pytest

from kopf import login


def test_client_login_works_incluster(login_mocks, kubernetes):
    with pytest.deprecated_call(match=r"cease using kopf.login\(\)"):
        login()

    assert login_mocks.client_in_cluster.called
    assert not login_mocks.client_from_file.called


def test_client_login_works_viaconfig(login_mocks, kubernetes):
    login_mocks.client_in_cluster.side_effect = kubernetes.config.ConfigException

    with pytest.deprecated_call(match=r"cease using kopf.login\(\)"):
        login()

    assert login_mocks.client_in_cluster.called
    assert login_mocks.client_from_file.called


def test_pykube_login_works_incluster(login_mocks, pykube):
    with pytest.deprecated_call(match=r"cease using kopf.login\(\)"):
        login()

    assert login_mocks.pykube_in_cluster.called
    assert not login_mocks.pykube_from_file.called


def test_pykube_login_works_viaconfig(login_mocks, pykube):
    login_mocks.pykube_in_cluster.side_effect = FileNotFoundError

    with pytest.deprecated_call(match=r"cease using kopf.login\(\)"):
        login()

    assert login_mocks.pykube_in_cluster.called
    assert login_mocks.pykube_from_file.called


def test_monkeypatched_get_pykube_cfg_overrides_pykube(mocker, login_mocks):
    get_pykube_cfg = mocker.patch('kopf.clients.auth.get_pykube_cfg')
    get_pykube_cfg.return_value = pykube.KubeConfig({
        'current-context': 'self',
        'contexts': [{'name': 'self', 'context': {'cluster': 'self'}}],
        'clusters': [{'name': 'self', 'cluster': {'server': 'https://localhost'}}],
    })

    with pytest.deprecated_call(match=r"cease using kopf.login\(\)"):
        login()

    assert get_pykube_cfg.called
    assert not login_mocks.pykube_in_cluster.called
    assert not login_mocks.pykube_from_file.called


def test_pykube_is_independent_of_client_incluster(login_mocks, no_kubernetes, pykube):
    with pytest.deprecated_call(match=r"cease using kopf.login\(\)"):
        login()

    assert login_mocks.pykube_in_cluster.called
    assert not login_mocks.pykube_from_file.called


def test_pykube_is_independent_of_client_viaconfig(login_mocks, no_kubernetes, pykube):
    login_mocks.pykube_in_cluster.side_effect = FileNotFoundError

    with pytest.deprecated_call(match=r"cease using kopf.login\(\)"):
        login()

    assert login_mocks.pykube_in_cluster.called
    assert login_mocks.pykube_from_file.called


def test_client_is_independent_of_pykube_incluster(login_mocks, no_pykube, kubernetes):
    with pytest.deprecated_call(match=r"cease using kopf.login\(\)"):
        login()

    assert login_mocks.client_in_cluster.called
    assert not login_mocks.client_from_file.called


def test_client_is_independent_of_pykube_viaconfig(login_mocks, no_pykube, kubernetes):
    login_mocks.client_in_cluster.side_effect = kubernetes.config.ConfigException

    with pytest.deprecated_call(match=r"cease using kopf.login\(\)"):
        login()

    assert login_mocks.client_in_cluster.called
    assert login_mocks.client_from_file.called



import asyncio
import dataclasses
import functools
from typing import Any, Callable, List, Optional

import click

from kopf.engines import loggers, peering
from kopf.reactor import registries, running
from kopf.structs import configuration, credentials, primitives
from kopf.utilities import loaders


@dataclasses.dataclass()
class CLIControls:
    """ `KopfRunner` controls, which are impossible to pass via CLI. """
    ready_flag: Optional[primitives.Flag] = None
    stop_flag: Optional[primitives.Flag] = None
    vault: Optional[credentials.Vault] = None
    registry: Optional[registries.OperatorRegistry] = None
    settings: Optional[configuration.OperatorSettings] = None


class LogFormatParamType(click.Choice):

    def __init__(self) -> None:
        super().__init__(choices=[v.name.lower() for v in loggers.LogFormat])

    def convert(self, value: Any, param: Any, ctx: Any) -> loggers.LogFormat:
        name: str = super().convert(value, param, ctx)
        return loggers.LogFormat[name.upper()]


def logging_options(fn: Callable[..., Any]) -> Callable[..., Any]:
    """ A decorator to configure logging in all commands the same way."""
    @click.option('-v', '--verbose', is_flag=True)
    @click.option('-d', '--debug', is_flag=True)
    @click.option('-q', '--quiet', is_flag=True)
    @click.option('--log-format', type=LogFormatParamType(), default='full')
    @click.option('--log-refkey', type=str)
    @click.option('--log-prefix/--no-log-prefix', default=None)
    @functools.wraps(fn)  # to preserve other opts/args
    def wrapper(verbose: bool, quiet: bool, debug: bool,
                log_format: loggers.LogFormat = loggers.LogFormat.FULL,
                log_prefix: Optional[bool] = False,
                log_refkey: Optional[str] = None,
                *args: Any, **kwargs: Any) -> Any:
        loggers.configure(debug=debug, verbose=verbose, quiet=quiet,
                          log_format=log_format, log_refkey=log_refkey, log_prefix=log_prefix)
        return fn(*args, **kwargs)

    return wrapper


@click.version_option(prog_name='kopf')
@click.group(name='kopf', context_settings=dict(
    auto_envvar_prefix='KOPF',
))
def main() -> None:
    pass


@main.command()
@logging_options
@click.option('-n', '--namespace', default=None)
@click.option('--standalone', is_flag=True, default=False)
@click.option('--dev', 'priority', type=int, is_flag=True, flag_value=666)
@click.option('-L', '--liveness', 'liveness_endpoint', type=str)
@click.option('-P', '--peering', 'peering_name', type=str, default=None, envvar='KOPF_RUN_PEERING')
@click.option('-p', '--priority', type=int, default=0)
@click.option('-m', '--module', 'modules', multiple=True)
@click.argument('paths', nargs=-1)
@click.make_pass_decorator(CLIControls, ensure=True)
def run(
        __controls: CLIControls,
        paths: List[str],
        modules: List[str],
        peering_name: Optional[str],
        priority: int,
        standalone: bool,
        namespace: Optional[str],
        liveness_endpoint: Optional[str],
) -> None:
    """ Start an operator process and handle all the requests. """
    if __controls.registry is not None:
        registries.set_default_registry(__controls.registry)
    loaders.preload(
        paths=paths,
        modules=modules,
    )
    return running.run(
        standalone=standalone,
        namespace=namespace,
        priority=priority,
        peering_name=peering_name,
        liveness_endpoint=liveness_endpoint,
        registry=__controls.registry,
        settings=__controls.settings,
        stop_flag=__controls.stop_flag,
        ready_flag=__controls.ready_flag,
        vault=__controls.vault,
    )


@main.command()
@logging_options
@click.option('-n', '--namespace', default=None)
@click.option('-i', '--id', type=str, default=None)
@click.option('--dev', 'priority', flag_value=666)
@click.option('-P', '--peering', 'peering_name', type=str, required=True, envvar='KOPF_FREEZE_PEERING')
@click.option('-p', '--priority', type=int, default=100, required=True)
@click.option('-t', '--lifetime', type=int, required=True)
@click.option('-m', '--message', type=str)
def freeze(
        id: Optional[str],
        message: Optional[str],
        lifetime: int,
        namespace: Optional[str],
        peering_name: str,
        priority: int,
) -> None:
    """ Freeze the resource handling in the cluster. """
    ourserlves = peering.Peer(
        id=id or peering.detect_own_id(),
        name=peering_name,
        namespace=namespace,
        priority=priority,
        lifetime=lifetime,
    )
    loop = asyncio.get_event_loop()
    loop.run_until_complete(ourserlves.keepalive())


@main.command()
@logging_options
@click.option('-n', '--namespace', default=None)
@click.option('-i', '--id', type=str, default=None)
@click.option('-P', '--peering', 'peering_name', type=str, required=True, envvar='KOPF_RESUME_PEERING')
def resume(
        id: Optional[str],
        namespace: Optional[str],
        peering_name: str,
) -> None:
    """ Resume the resource handling in the cluster. """
    ourselves = peering.Peer(
        id=id or peering.detect_own_id(),
        name=peering_name,
        namespace=namespace,
    )
    loop = asyncio.get_event_loop()
    loop.run_until_complete(ourselves.disappear())
